from pystage.core._base_sprite import BaseSprite

class _Control(BaseSprite):

    def __init__(self):
        super().__init__()

    def control_wait(self, secs):
        self.code_manager.current_block.add_to_wait_time = secs

    def control_stop_all(self):
        for s in self.stage.sprites:
            s.code_manager.stop_running_blocks()
        self.stage.code_manager.stop_running_blocks()

    control_stop_all.opcode="control_stop"
    control_stop_all.param="STOP_OPTION"
    control_stop_all.value="all"

    def control_stop_this(self):
        self.code_manager.stop_running_blocks()

    control_stop_this.opcode="control_stop"
    control_stop_this.param="STOP_OPTION"
    control_stop_this.value="this script"

    def control_stop_other(self):
        for s in self.stage.sprites:
            if s is not self:
                s.code_manager.stop_running_blocks()
        self.stage.code_manager.stop_running_blocks()

    control_stop_other.opcode="control_stop"
    control_stop_other.param="STOP_OPTION"
    control_stop_other.value="other scripts in sprite"

    # Cloning is probably tricky.  
    def control_create_clone_of_sprite(self, sprite):
        print("Creating a clone of ", sprite)
        sprite._core.clone()

    control_create_clone_of_sprite.opcode = "control_create_clone_of"

    
class _ControlSprite(_Control):

    number_of_clones = 0

    def __init__(self):
        super().__init__()

    def control_create_clone_of_myself(self):
        print("Creating a clone of myself")
        self._core.clone()

    control_create_clone_of_myself.opcode="control_create_clone_of"
    control_create_clone_of_myself.param="CLONE_OPTION"
    control_create_clone_of_myself.value="_myself_"

    def control_delete_this_clone(self):
        _ControlSprite.number_of_clones -= 1
        self.kill()
        self.code_manager.stop_running_blocks()
    control_delete_this_clone.translation = "control_deletethisclone"


    # This is actually an event but Scratch has the hat block under "Control"
    def control_start_as_clone(self, generator_function, name="", no_refresh=False):
        '''
        Adds the code block to the event queue for clicks.
        '''
        new_block = self.code_manager.register_code_block(generator_function, name, no_refresh)
        self.code_manager.cloned_blocks.append(new_block)
        print(f"Bound to start as clone: {new_block.name}")
    control_start_as_clone.translation = "control_startasclone"

    def clone(self):
        # In Scratch, there can only be a maximum of 300 Clones
        if _ControlSprite.number_of_clones >= 300:
            return
        _ControlSprite.number_of_clones += 1
        clone = self.stage._core.pystage_createsprite()._core
        clone.is_clone = True
        # All cloned by reference, be careful where this matters
        cloned_attributes = [
            # Motion
            "_direction", "draggable", 
            # Pen
            "pen", "pen_color", "pen_size", "old_position", "pen_up_at",
            # Looks
            "size", "ghost", "visible",
            # Sound
            "sound_manager", "current_pan", "current_pitch", "current_volume", 
            ]
        for attr in cloned_attributes:
            setattr(clone, attr, getattr(self, attr))
        # copy position over
        clone._pos.x = self._pos.x
        clone._pos.y = self._pos.y
        # clones have own variables 
        clone.variables = self.variables.copy()
        # But only the original variable can be monitored.
        # However, monitors can be shown and hidden, so we need access
        clone.monitors = self.monitors
        
        # Costumes
        clone.costume_manager.costumes = self.costume_manager.costumes
        clone.costume_manager.current_costume = self.costume_manager.current_costume
        clone.costume_manager.rotation_style = self.costume_manager.rotation_style

        # Code
        clone.code_manager = self.code_manager.clone(clone)
        for code_block in clone.code_manager.cloned_blocks:
            code_block.start_or_restart()


# pyright: basic, reportPrivateUsage=false
import sys

import pygame

# Mixins
from pystage.core._control import _Control
from pystage.core._events import _Events
from pystage.core._looks import _LooksStage
from pystage.core._operators import _Operators
from pystage.core._sensing import _Sensing
from pystage.core._sound import _Sound
from pystage.core._variables import _Variables
from pystage.core.messages import MessageBroker
from pystage.core.sprite import CoreSprite
from pystage.core.asking import InputManager


class SpriteGroup(pygame.sprite.OrderedUpdates):
    def __init__(self):
        super().__init__()
        # Be careful, we access a private variable here
        self._spritelist = self._spritelist

    def get_layer(self, sprite):
        try:
            return self.sprites().index(sprite)
        except ValueError:
            return -1

    def to_front(self, sprite):
        if not sprite in self._spritelist:
            raise ValueError(f"Sprite {sprite} unknown.")
        self._spritelist.remove(sprite)
        self._spritelist.append(sprite)
        # print(self._spritelist)

    def to_back(self, sprite):
        if not sprite in self._spritelist:
            raise ValueError(f"Sprite {sprite} unknown.")
        self._spritelist.remove(sprite)
        self._spritelist.insert(0, sprite)
        # print(self._spritelist)

    def layer_forward(self, sprite, value=1):
        if not sprite in self._spritelist:
            raise ValueError(f"Sprite {sprite} unknown.")
        index = self._spritelist.index(sprite)
        self._spritelist.remove(sprite)
        new_index = index + value
        if new_index < 0:
            new_index = 0
        # because the sprite was already removed above, so don't need to minus 1 for the length
        if new_index > len(self._spritelist):
            new_index = len(self._spritelist)
        self._spritelist.insert(new_index, sprite)
        # print(self._spritelist)

    def layer_backward(self, sprite, value=1):
        self.layer_forward(sprite, -value)

    def insert(self, i, sprite):
        self._spritelist.insert(i, sprite)


class CoreStage(
    _LooksStage, _Sound, _Events, _Control, _Operators, _Sensing, _Variables
):
    def __init__(self, name="Welcome to pyStage!", width=480, height=360):
        # Check pygame version
        pygame_major = int(pygame.ver.split(".")[0])
        if pygame_major < 2:
            print("pygame version 2 or higher is required for PyStage.")
            print("Update with the following command:")
            print("pip install --upgrade pygame")
            sys.exit(1)

        # This way, code blocks can consistently refer to the stage with self.stage:
        self.stage = self
        # Make sure that the core sprite can always be accessed via _core
        self._core = self
        # Above attributes need to be set first so that mixins can access them properly
        super().__init__()
        # The facade is the translated API
        self.facade = None
        self.sprite_facade_class = None

        self.message_broker = MessageBroker(self)
        self.input_manager = InputManager(self)

        pygame.init()
        pygame.display.set_caption(name)
        self.running = False
        self.FPS = 60
        self.dt = 0
        # The stage is added to the sprites as it also contains code.
        self.sprites = SpriteGroup()
        self.visible_sprites = SpriteGroup()
        self.bubbles = pygame.sprite.Group()
        self.visible_bubbles = pygame.sprite.Group()
        self.monitor_group = pygame.sprite.Group()
        self.pen_images = {}
        self.background_color = (255, 255, 255)
        # surface is where the whole stage is rendered to
        # it defines the in-game resolution
        self.surface = pygame.Surface([width, height], flags=pygame.SRCALPHA)
        # screen is the actual screen, the surface gets scaled
        self.screen = pygame.display.set_mode([width, height], pygame.RESIZABLE)
        self.clock = pygame.time.Clock()
        self.width = width
        self.height = height
        self.rect = pygame.rect.Rect(0, 0, width, height)
        self.center_x = int(width / 2)
        self.center_y = int(height / 2)
        # current scale factor of the stage if the window is resized
        self.scale_factor = 1
        # current offset of the top left corner if the window is resized
        self.offset_x = 0
        self.offset_y = 0

        self.timer = 0
        self.show_sprite_boundaries = "--show-sprite-boundaries" in sys.argv

    def pystage_createsprite(self, costume="default"):
        sprite = CoreSprite(self, costume)
        self.sprites.add(sprite)
        self._update_visible()
        if self.sprite_facade_class:
            return self.sprite_facade_class(sprite)  # pylint: disable=E1102
            # pylint produces a false positive here
            # https://github.com/PyCQA/pylint/issues/1493
        else:
            return sprite

    def _update_visible(self):
        self.visible_sprites.empty()
        self.visible_bubbles.empty()
        for sprite in self.sprites:
            assert(isinstance(sprite, CoreSprite))
            if sprite.visible:
                self.visible_sprites.add(sprite)
                if sprite.bubble_manager.bubble:
                    self.visible_bubbles.add(sprite.bubble_manager.bubble)

    def _update(self, dt):
        self.code_manager._update(dt)
        # Update it so that the layering is correct while drawing
        # otherwise the no effect for to_front, to_back, etc.
        self._update_visible()

    def _draw(self, surface: pygame.Surface):
        surface.fill(self.background_color)
        image = self.costume_manager.get_image()
        if not image:
            return
        surface.blit(image, (0, 0))

    def pystage_play(self):
        self.running = True
        """
        This runs the game loop
        """
        dt = 0

        # maybe we should move this to other place
        left_clicking_and_holding = False
        clicked_sprite = None

        while self.running:
            for event in pygame.event.get():
                if event.type == pygame.QUIT:
                    self.running = False
                if event.type == pygame.KEYDOWN:
                    if self.input_manager.is_active():
                        self.input_manager.process_key(event)
                    else:
                        for sprite in self.sprites:
                            assert(isinstance(sprite, CoreSprite))
                            sprite.code_manager.process_key_pressed(event.key)
                # to make it consistent with Scratch, should be BUTTONDOWN
                if event.type == pygame.MOUSEBUTTONDOWN:
                    pos = pygame.Vector2(pygame.mouse.get_pos())
                    # [-1:0:-1] will reverse the list, but not include 0
                    for sprite in self.visible_sprites.sprites()[-1::-1]:
                        assert(isinstance(sprite, CoreSprite))
                        if sprite.rect.collidepoint(pos):
                            internal_pos = pos - sprite.rect.topleft
                            x = int(internal_pos.x)
                            y = int(internal_pos.y)
                            color = sprite.image.get_at((x, y))
                            if color.a == 0:
                                continue
                            sprite.code_manager.process_click()
                            break

                # with these three conditions, we can detect a click and hold, 
                # but not move the mouse while holding
                if event.type == pygame.MOUSEBUTTONDOWN and not left_clicking_and_holding and pygame.mouse.get_pressed()[0]:
                    pos = pygame.mouse.get_pos()
                    for sprite in filter(lambda s:s.draggable, self.visible_sprites.sprites()[-1::-1]):
                        assert(isinstance(sprite, CoreSprite))
                        sprite_rect = sprite.image.get_rect()
                        sprite_rect.topleft = sprite.rect.topleft
                        if sprite_rect.collidepoint(pos):
                            clicked_pos = pos[0] - sprite_rect.left, pos[1] - sprite_rect.top
                            if sprite.image.get_at(clicked_pos).a != 0:
                                left_clicking_and_holding = True
                                clicked_sprite = sprite

            if not pygame.mouse.get_pressed()[0] and left_clicking_and_holding:
                left_clicking_and_holding = False

            if left_clicking_and_holding:
                assert isinstance(clicked_sprite, CoreSprite)
                clicked_sprite.motion_goto_pointer()

            # Handle broadcast messages
            for message in self.message_broker.get_messages():
                for sprite in self.sprites:
                    assert(isinstance(sprite, CoreSprite))
                    sprite.code_manager.process_broadcast(message)
            self.message_broker.mark_completed()

            self._update(dt)
            self.sprites.update(dt)
            self.bubbles.update()
            self.input_manager.update(dt)
            self.monitor_group.update()

            self._draw(self.surface)
            for sprite in self.pen_images:
                image = self.pen_images[sprite]
                self.surface.blit(image, (0, 0))

            self.visible_sprites.draw(self.surface)
            self.bubbles.draw(self.surface)
            self.input_manager.draw(self.surface)

            self.monitor_group.draw(self.surface)

            factor_x = self.screen.get_width() / self.surface.get_width()
            factor_y = self.screen.get_height() / self.surface.get_height()
            self.scale_factor = min(factor_x, factor_y)
            scaled = pygame.transform.smoothscale(
                self.surface,
                (
                    int(self.surface.get_width() * self.scale_factor),
                    int(self.surface.get_height() * self.scale_factor),
                ),
            )
            self.offset_x = int((self.screen.get_width() - scaled.get_width()) / 2)
            self.offset_y = int((self.screen.get_height() - scaled.get_height()) / 2)
            self.screen.blit(scaled, (self.offset_x, self.offset_y,))
            pygame.display.flip()

            dt = self.clock.tick(self.FPS) / 1000
            self.timer += dt

        pygame.quit()



import pygame
import pystage
from pystage.core.constants import KEY_MAPPINGS
from pystage.core._base_sprite import BaseSprite
from math import sqrt
import datetime


class _Sensing(BaseSprite):

    def __init__(self):
        super().__init__()


    def sensing_askandwait(self, question):
        self.code_manager.current_block.ask(question)

    def sensing_answer(self):
        return self.stage.input_manager.answer

    def sensing_keypressed(self, key):
        return pygame.key.get_pressed()[KEY_MAPPINGS[key]]   

    def sensing_mousedown(self):
        return any(pygame.mouse.get_pressed())

    def sensing_mousex(self):
        x = pygame.mouse.get_pos()[0]
        return ((x - self.stage.offset_x) / self.stage.scale_factor) - self.stage.width / 2

    def sensing_mousey(self):
        y = pygame.mouse.get_pos()[1]
        return -(y - self.stage.offset_y) / self.stage.scale_factor + self.stage.height / 2

    def sensing_loudness(self):
        # See events area, not sure if we support microphone access
        # This could be useful: https://python-sounddevice.readthedocs.io/en/0.4.7/index.html
        print("Microphone access is currently not supported. See https://github.com/pystage/pystage/issues/116")

    def sensing_timer(self):
        return self.stage.timer

    def sensing_resettimer(self):
        self.stage.timer = 0


    def sensing_setdragmode_draggable(self):
        self.draggable = True
    sensing_setdragmode_draggable.opcode="sensing_setdragmode"
    sensing_setdragmode_draggable.param="DRAG_MODE"
    sensing_setdragmode_draggable.value="draggable"


    def sensing_setdragmode_notdraggable(self):
        self.draggable = False
    sensing_setdragmode_notdraggable.opcode="sensing_setdragmode"
    sensing_setdragmode_notdraggable.param="DRAG_MODE"
    sensing_setdragmode_notdraggable.value="not draggable"

    # Here follows the mess of our favorite block, where a lot of stuff 
    # from other sprites and the stage can be retrieved.
    # This is all redundant, as in Python you would simply do something like
    # self.stage.get_backdrop_name(). Along these lines, we need good
    # explanations and examples in the documentation that show why these functions
    # should not be used.

    def sensing_of_xposition(self, sprite):
        pass
    sensing_of_xposition.opcode="sensing_of"
    sensing_of_xposition.param="PROPERTY"
    sensing_of_xposition.value="x position"


    def sensing_of_yposition(self, sprite):
        pass
    sensing_of_yposition.opcode="sensing_of"
    sensing_of_yposition.param="PROPERTY"
    sensing_of_yposition.value="y position"


    def sensing_of_direction(self, sprite):
        pass
    sensing_of_direction.opcode="sensing_of"
    sensing_of_direction.param="PROPERTY"
    sensing_of_direction.value="direction"


    def sensing_of_costumenumber(self, sprite):
        pass
    sensing_of_costumenumber.opcode="sensing_of"
    sensing_of_costumenumber.param="PROPERTY"
    sensing_of_costumenumber.value="costume #"


    def sensing_of_costumename(self, sprite):
        pass
    sensing_of_costumename.opcode="sensing_of"
    sensing_of_costumename.param="PROPERTY"
    sensing_of_costumename.value="costume name"


    def sensing_of_size(self, sprite):
        pass
    sensing_of_size.opcode="sensing_of"
    sensing_of_size.param="PROPERTY"
    sensing_of_size.value="size"


    def sensing_of_volume(self, sprite="_stage_"):
        pass
    sensing_of_volume.opcode="sensing_of"
    sensing_of_volume.param="PROPERTY"
    sensing_of_volume.value="volume"


    def sensing_of_variable(self, variable, sprite="_stage_"):
        pass
    sensing_of_variable.opcode="sensing_of"


    def sensing_of_backdropnumber(self, stage="_stage_"):
        pass
    sensing_of_backdropnumber.opcode="sensing_of"
    sensing_of_backdropnumber.param="PROPERTY"
    sensing_of_backdropnumber.value="backdrop #"


    def sensing_of_backdropname(self, stage="_stage_"):
        pass
    sensing_of_backdropname.opcode="sensing_of"
    sensing_of_backdropname.param="PROPERTY"
    sensing_of_backdropname.value="backdrop name"

    def sensing_current_year(self):
        return datetime.datetime.now().year
    sensing_current_year.opcode="sensing_current"
    sensing_current_year.param="CURRENTMENU"
    sensing_current_year.value="YEAR"


    def sensing_current_month(self):
        return datetime.datetime.now().month
    sensing_current_month.opcode="sensing_current"
    sensing_current_month.param="CURRENTMENU"
    sensing_current_month.value="MONTH"


    def sensing_current_date(self):
        return datetime.datetime.now().day
    sensing_current_date.opcode="sensing_current"
    sensing_current_date.param="CURRENTMENU"
    sensing_current_date.value="DATE"


    def sensing_current_dayofweek(self):
        current_datetime = datetime.datetime.now().date()
        #This leads to a Sunday=1 based counting
        day_of_week = (current_datetime.isoweekday() % 7) + 1
        return day_of_week
    sensing_current_dayofweek.opcode="sensing_current"
    sensing_current_dayofweek.param="CURRENTMENU"
    sensing_current_dayofweek.value="DAYOFWEEK"


    def sensing_current_hour(self):
        return datetime.datetime.now().hour
    sensing_current_hour.opcode="sensing_current"
    sensing_current_hour.param="CURRENTMENU"
    sensing_current_hour.value="HOUR"


    def sensing_current_minute(self):
        return datetime.datetime.now().minute
    sensing_current_minute.opcode="sensing_current"
    sensing_current_minute.param="CURRENTMENU"
    sensing_current_minute.value="MINUTE"


    def sensing_current_second(self):
        return datetime.datetime.now().second
    sensing_current_second.opcode="sensing_current"
    sensing_current_second.param="CURRENTMENU"
    sensing_current_second.value="SECOND"


    def sensing_dayssince2000(self):
        # use datetime to find days since 2000 in timezone
        start_date = datetime.datetime(2000, 1, 1, tzinfo=datetime.timezone.utc)
        current_datetime = datetime.datetime.now(datetime.timezone.utc)
        time_difference = current_datetime - start_date
        days = time_difference.days
        # find fraction of current day like in Scratch
        decimal_in_day_string = (current_datetime.hour * 3600 + current_datetime.minute * 60 + current_datetime.second) / (24 * 3600)
        # put full days and part of day together
        since_2000 = days + decimal_in_day_string
        return since_2000

    def sensing_username(self):
        # Makes not a lot of sense, maybe for compatibility?
        pass

class _SensingSprite(BaseSprite):
    def __init__(self):
        super().__init__()

    def sensing_touchingobject_pointer(self):
        pos = pygame.mouse.get_pos()
        if self.rect.collidepoint(pos):
            x = pos[0] - self.rect.left
            y = pos[1] - self.rect.top
            color = self.image.get_at((x, y))
            if color.a != 0:
                return True
        return False

    sensing_touchingobject_pointer.opcode="sensing_touchingobject"
    sensing_touchingobject_pointer.param="TOUCHINGOBJECTMENU"
    sensing_touchingobject_pointer.value="_mouse_"

    def sensing_touchingobject_edge(self):
        return not self.stage.rect.contains(self.rect)

    sensing_touchingobject_edge.opcode="sensing_touchingobject"
    sensing_touchingobject_edge.param="TOUCHINGOBJECTMENU"
    sensing_touchingobject_edge.value="_edge_"

    def sensing_touchingobject_sprite(self, sprite):
        sprite = sprite._core
        if sprite.rect.colliderect(self.rect):
            offset = (self.rect.left - sprite.rect.left, self.rect.top - sprite.rect.top)
            return sprite.mask.overlap(self.mask, offset) is not None
        return False

    sensing_touchingobject_sprite.opcode="sensing_touchingobject"

    def sensing_touchingcolor(self, color):
        pass

    def sensing_coloristouchingcolor(self, sprite_color, color):
        pass

    def sensing_distanceto_pointer(self):
        mx, my = pygame.mouse.get_pos()
        # gets position of sprite
        sprite_rect = self.rect
        # Calculate the distance between the sprite and the mouse pointer
        distance = sqrt((sprite_rect.centerx - mx)**2 + (sprite_rect.centery - my)**2)
        return distance
        
    sensing_distanceto_pointer.opcode="sensing_distanceto"
    sensing_distanceto_pointer.param="DISTANCETOMENU"
    sensing_distanceto_pointer.value="_mouse_"

    def sensing_distanceto_sprite(self, sprite):
        # self is an instance of CoreSprite
        # sprite is an instance of Sprite
        return self._pos.distance_to(sprite._core._pos)
    sensing_distanceto_sprite.opcode="sensing_distanceto"





import pygame
from pystage.core._base_sprite import BaseSprite



def int2(vec2):
    x = int(round(vec2[0]))
    y = int(round(vec2[1]))
    return pygame.Vector2(x, y)


class _Pen(BaseSprite):

    def __init__(self):
        super().__init__()

        self.pen = False
        self.pen_color = pygame.Color(255,0,0)
        self.pen_size = 3
        self.old_position = (0,0)
        self.pen_up_at = (0,0)


    def _get_image(self):
        if not self in self.stage.pen_images:
            image = pygame.Surface((self.stage.width, self.stage.height), flags=pygame.SRCALPHA)
            self.stage.pen_images[self] = image
        return self.stage.pen_images[self]

    def _update_pen(self):
        if self.pen:
            if self.old_position == self._pos:
                return
            pygame.draw.line(self._get_image(), self.pen_color, self.old_position, self._pos, width=int(self.pen_size))
            self.old_position = list(self._pos)
        else:
            if self.pen_up_at != self.old_position:
                pygame.draw.line(self._get_image(), self.pen_color, self.old_position, self.pen_up_at, width=int(self.pen_size))
                self.pen_up_at = self.old_position

    def pen_clear(self):
        if self in self.stage.pen_images:
            del self.stage.pen_images[self]


    def pen_stamp(self):
        pass


    def pen_penDown(self):
        if not self.pen:
            self.old_position = list(self._pos)
        self.pen = True


    def pen_penUp(self):
        self.pen_up_at = self._pos
        self.pen = False


    def pen_setPenColorToColor(self, color):
        self.pen_color = pygame.Color(color)

    pen_setPenColorToColor.translation = "pen_setcolor"


    def pen_changePenColorParamBy_color(self, value):
        hsva = list(self.pen_color.hsva)
        hsva[0] = (hsva[0] + (value * 3.6)) % 360
        self.pen_color.hsva = hsva

    pen_changePenColorParamBy_color.opcode="pen_changePenColorParamBy"
    pen_changePenColorParamBy_color.param="COLOR_PARAM"
    pen_changePenColorParamBy_color.value="color"
    pen_changePenColorParamBy_color.translation="pen_colormenu_color"
    pen_changePenColorParamBy_color.translation_hint="Needs to be distinguished from setPenColorToColor. This is only the hue value."
    pen_changePenColorParamBy_color.outer_translation="pen_changecolorparam"
    pen_changePenColorParamBy_color.position="[COLOR_PARAM]"


    def pen_changePenColorParamBy_saturation(self, value):
        hsva = list(self.pen_color.hsva)
        hsva[1] = (hsva[1] + (value)) % 100
        self.pen_color.hsva = hsva

    pen_changePenColorParamBy_saturation.opcode="pen_changePenColorParamBy"
    pen_changePenColorParamBy_saturation.param="COLOR_PARAM"
    pen_changePenColorParamBy_saturation.value="saturation"
    pen_changePenColorParamBy_saturation.translation="pen_colormenu_saturation"
    pen_changePenColorParamBy_saturation.outer_translation="pen_changecolorparam"
    pen_changePenColorParamBy_saturation.position="[COLOR_PARAM]"
    

    def pen_changePenColorParamBy_brightness(self, value):
        hsva = list(self.pen_color.hsva)
        hsva[2] = (hsva[2] + (value)) % 100
        self.pen_color.hsva = hsva

    pen_changePenColorParamBy_brightness.opcode="pen_changePenColorParamBy"
    pen_changePenColorParamBy_brightness.param="COLOR_PARAM"
    pen_changePenColorParamBy_brightness.value="brightness"
    pen_changePenColorParamBy_brightness.translation="pen_colormenu_brightness"
    pen_changePenColorParamBy_brightness.outer_translation="pen_changecolorparam"
    pen_changePenColorParamBy_brightness.position="[COLOR_PARAM]"
    

    def pen_changePenColorParamBy_transparency(self, value):
        hsva = list(self.pen_color.hsva)
        hsva[3] = 100 - (((100 - hsva[3]) + (value)) % 100)
        self.pen_color.hsva = hsva

    pen_changePenColorParamBy_transparency.opcode="pen_changePenColorParamBy"
    pen_changePenColorParamBy_transparency.param="COLOR_PARAM"
    pen_changePenColorParamBy_transparency.value="transparency"
    pen_changePenColorParamBy_transparency.translation="pen_colormenu_transparency"
    pen_changePenColorParamBy_transparency.outer_translation="pen_changecolorparam"
    pen_changePenColorParamBy_transparency.position="[COLOR_PARAM]"


    def pen_setPenColorParamTo_color(self, value):
        hsva = list(self.pen_color.hsva)
        hsva[0] = (value * 3.6) % 360 
        self.pen_color.hsva = hsva

    pen_setPenColorParamTo_color.opcode="pen_setPenColorParamTo"
    pen_setPenColorParamTo_color.param="COLOR_PARAM"
    pen_setPenColorParamTo_color.value="color"
    pen_setPenColorParamTo_color.translation="pen_colormenu_color"
    pen_setPenColorParamTo_color.translation_hint="Needs to be distinguished from setPenColorToColor. This is only the hue value."
    pen_setPenColorParamTo_color.outer_translation="pen_setcolorparam"
    pen_setPenColorParamTo_color.position="[COLOR_PARAM]"


    def pen_setPenColorParamTo_saturation(self, value):
        hsva = list(self.pen_color.hsva)
        hsva[1] = value % 100
        self.pen_color.hsva = hsva

    pen_setPenColorParamTo_saturation.opcode="pen_setPenColorParamTo"
    pen_setPenColorParamTo_saturation.param="COLOR_PARAM"
    pen_setPenColorParamTo_saturation.value="saturation"
    pen_setPenColorParamTo_saturation.translation="pen_colormenu_saturation"
    pen_setPenColorParamTo_saturation.outer_translation="pen_setcolorparam"
    pen_setPenColorParamTo_saturation.position="[COLOR_PARAM]"
    

    def pen_setPenColorParamTo_brightness(self, value):
        hsva = list(self.pen_color.hsva)
        hsva[2] = value % 100
        self.pen_color.hsva = hsva

    pen_setPenColorParamTo_brightness.opcode="pen_setPenColorParamTo"
    pen_setPenColorParamTo_brightness.param="COLOR_PARAM"
    pen_setPenColorParamTo_brightness.value="brightness"
    pen_setPenColorParamTo_brightness.translation="pen_colormenu_brightness"
    pen_setPenColorParamTo_brightness.outer_translation="pen_setcolorparam"
    pen_setPenColorParamTo_brightness.position="[COLOR_PARAM]"
    

    def pen_setPenColorParamTo_transparency(self, value):
        hsva = list(self.pen_color.hsva)
        hsva[3] = 100 - (value % 100)
        self.pen_color.hsva = hsva

    pen_setPenColorParamTo_transparency.opcode="pen_setPenColorParamTo"
    pen_setPenColorParamTo_transparency.param="COLOR_PARAM"
    pen_setPenColorParamTo_transparency.value="transparency"
    pen_setPenColorParamTo_transparency.translation="pen_colormenu_transparency"
    pen_setPenColorParamTo_transparency.outer_translation="pen_setcolorparam"
    pen_setPenColorParamTo_transparency.position="[COLOR_PARAM]"


    def pen_changePenSizeBy(self, value):
        self.pen_size += value

    pen_changePenSizeBy.translation = "pen_changesize"

    def pen_setPenSizeTo(self, value):
        self.pen_size = value

    pen_setPenSizeTo.translation = "pen_setsize"



import pygame
import pkg_resources

pygame.init()


regular_font_14 = pygame.font.Font(pkg_resources.resource_filename("pystage", "fonts/roboto-regular.ttf"), 14)
bold_font_9 = pygame.font.Font(pkg_resources.resource_filename("pystage", "fonts/roboto-bold.ttf"), 9)
light_font_9 = pygame.font.Font(pkg_resources.resource_filename("pystage", "fonts/roboto-bold.ttf"), 9)

color1 = (87,94,117) # Text in bubbles
# Variable name bold 9
# Variable value light 9
# bubble text regular 14

class ResizableBorder():

    def __init__(self, image_file, inner_rect):
        self.image = pygame.image.load(image_file)
        self.inner_rect = inner_rect
        w = self.image.get_width()
        h = self.image.get_height()
        ix, iy, iw, ih = inner_rect
        self.ul = pygame.Surface((ix, iy), flags=pygame.SRCALPHA)
        self.top = pygame.Surface((iw, iy), flags=pygame.SRCALPHA)
        self.ur = pygame.Surface((w - (ix + iw), iy), flags=pygame.SRCALPHA)
        self.left = pygame.Surface((ix, ih), flags=pygame.SRCALPHA)
        self.center = pygame.Surface((iw, ih), flags=pygame.SRCALPHA)
        self.right = pygame.Surface((w - (ix + iw), ih), flags=pygame.SRCALPHA)
        self.bl = pygame.Surface((ix, h - (iy + ih)), flags=pygame.SRCALPHA)
        self.bottom = pygame.Surface((iw, h - (iy + ih)), flags=pygame.SRCALPHA)
        self.br = pygame.Surface((w - (ix + iw), h - (iy + ih)), flags=pygame.SRCALPHA)
        self.ul.blit(self.image, (0,0), (0, 0, self.ul.get_width(), self.ul.get_height()))
        self.top.blit(self.image, (0,0), (ix, 0, self.top.get_width(), self.top.get_height()))
        self.ur.blit(self.image, (0,0), (ix + iw, 0, self.ur.get_width(), self.ur.get_height()))
        self.left.blit(self.image, (0,0), (0, iy, self.left.get_width(), self.left.get_height()))
        self.center.blit(self.image, (0,0), (ix, iy, self.center.get_width(), self.center.get_height()))
        self.right.blit(self.image, (0,0), (ix + iw, iy, self.right.get_width(), self.right.get_height()))
        self.bl.blit(self.image, (0,0), (0, iy + ih, self.bl.get_width(), self.bl.get_height()))
        self.bottom.blit(self.image, (0,0), (ix, iy + ih, self.bottom.get_width(), self.bottom.get_height()))
        self.br.blit(self.image, (0,0), (ix + iw, iy + ih, self.br.get_width(), self.br.get_height()))
        self.min_width = self.ul.get_width() + self.ur.get_width()
        self.min_height = self.ul.get_height() + self.bl.get_height()


    def render(self, width, height):
        iw = max(width - self.min_width, 0)
        ih = max(height - self.min_height, 0)
        surface = pygame.Surface((self.min_width + iw, self.min_height + ih), flags=pygame.SRCALPHA)

        scaled_top = pygame.transform.scale(self.top, (iw, self.top.get_height()))
        surface.blit(scaled_top, (self.ul.get_width(), 0))
        scaled_bottom = pygame.transform.scale(self.bottom, (iw, self.bottom.get_height()))
        surface.blit(scaled_bottom, (self.ul.get_width(), self.ul.get_height() + ih))
        scaled_left = pygame.transform.scale(self.left, (self.left.get_width(), ih))
        surface.blit(scaled_left, (0, self.ul.get_height()))
        scaled_right = pygame.transform.scale(self.right, (self.right.get_width(), ih))
        surface.blit(scaled_right, (self.left.get_width() + iw, self.ul.get_height()))

        scaled_center = pygame.transform.scale(self.center, (iw, ih))
        surface.blit(scaled_center, (self.ul.get_width(), self.ul.get_height()))

        surface.blit(self.ul, (0, 0))
        surface.blit(self.ur, (self.ul.get_width() + iw, 0))
        surface.blit(self.bl, (0, self.ul.get_height() + ih))
        surface.blit(self.br, (self.ul.get_width() + iw, self.ul.get_height() + ih))

        return surface


def wrap_text(text, font, allowed_width):
    # See: https://stackoverflow.com/questions/49432109/how-to-wrap-text-in-pygame-using-pygame-font-font
    # first, split the text into words
    words = text.split()

    # now, construct lines out of these words
    lines = []
    max_lw = 0
    max_lh = 0
    while len(words) > 0:
        # get as many words as will fit within allowed_width
        line_words = []
        while len(words) > 0:
            line_words.append(words.pop(0))
            if len(line_words)==1:
                # Check for long words
                lw, lh = font.size(line_words[0])
                max_lh = lh
                max_lw = max(max_lw, lw)
                if lw > allowed_width:
                    max_lw = allowed_width
                    cut = []
                    while lw > allowed_width:
                        cut.append(line_words[0][-1]) 
                        line_words[0] = line_words[0][:-1]
                        lw, lh = font.size(line_words[0])
                    cut.reverse()
                    words.insert(0, "".join(cut))
                    break
            else:
                lw, lh = font.size(' '.join(line_words + words[:1]))
                if lw > allowed_width:
                    break
                max_lw = max(max_lw, lw)


        # add a line consisting of those words
        line = ' '.join(line_words)
        lines.append(line)
        # print(lines, max_lw, max_lh)
    return lines, max_lw, max_lh


def render_lines(lines, font, color, lw, lh, lh_offset):
    # we'll render each line below the last, so we need to keep track of
    # the culmative height of the lines we've rendered so far
    surface = pygame.Surface((lw, (lh + lh_offset) * len(lines)), flags=pygame.SRCALPHA)
    y_offset = 0
    for line in lines:
        font_surface = font.render(line, True, color)
        surface.blit(font_surface, (0, y_offset))
        y_offset += lh + lh_offset
    return surface



class Bubble(pygame.sprite.Sprite):
    say_file = pkg_resources.resource_filename("pystage", "images/say.png")
    think_file = pkg_resources.resource_filename("pystage", "images/think.png")
    
    SAY = ResizableBorder(say_file, (18,18,154, 35))
    THINK = ResizableBorder(think_file, (18,18,154, 35))

    def __init__(self, sprite, text: str, border: ResizableBorder):
        super().__init__()
        self.sprite = sprite
        self.text = text
        self.border = border
        self.flipped = False
        self.sprite_x = self.sprite.motion_xposition()
        self.sprite_y = self.sprite.motion_yposition()
        self.x_offset = -0.2 # percent of sprite width based on lower right corner
        self.y_offset = -0.2 # percent of sprite height based on lower right corner
        lh_offset = -1
        lines, lw, lh = wrap_text(self.text[:360], regular_font_14, 170)
        self.text_surface = render_lines(lines, regular_font_14, color1, lw+10, lh, lh_offset)
        self.image_normal = self.border.render(self.text_surface.get_width() + 26, self.text_surface.get_height() + 40)
        self.image_flipped = pygame.transform.flip(self.image_normal, True, False)
        self.image_normal.blit(self.text_surface, (13, 13))
        self.image_flipped.blit(self.text_surface, (13, 13))
        self.image = self.image_normal
        self.rect = self.image.get_rect()
        self.update(force=True)


    def update(self, force=False):
        if not force and self.sprite.motion_xposition() == self.sprite_x and self.sprite.motion_yposition() == self.sprite_y:
            return
        # update position to avoid repeated updates
        self.sprite_x = self.sprite.motion_xposition()
        self.sprite_y = self.sprite.motion_yposition()
        y = max(0, self.sprite.rect.top - self.image_normal.get_height() - self.sprite.costume_manager.get_image().get_height() * self.y_offset)
        if not self.flipped:
            x = self.sprite.rect.left - self.image_normal.get_width() - self.sprite.costume_manager.get_image().get_width() * self.x_offset
            if x < 0:
                self.flipped = True
                x = self.sprite.rect.right + self.sprite.costume_manager.get_image().get_width() * self.x_offset
        else:
            x = self.sprite.rect.right + self.sprite.costume_manager.get_image().get_width() * self.x_offset
            if x + self.image_normal.get_width() > self.sprite.stage.width:
                self.flipped = False
                x = self.sprite.rect.left - self.image_normal.get_width() - self.sprite.costume_manager.get_image().get_width() * self.x_offset
        self.image = self.image_flipped if self.flipped else self.image_normal
        self.rect.x = x
        self.rect.y = y




class BubbleManager():

    def __init__(self, sprite):
        self.sprite = sprite
        self.flipped = False
        self.bubble = None

    def say(self, text: str, border=Bubble.SAY):
        # print(text)
        if isinstance(text, (int, float)):
            # trim the non-zero fractional part
            # e.g. 8.00 will be 8
            text = str(round(text, 2) if text % 1 > 0 else int(text))
        if self.bubble:
            self.bubble.kill()
            self.bubble = None
        if text is not None and text.strip() != "":
            self.bubble = Bubble(self.sprite, text, border)
            self.sprite.stage.bubbles.add(self.bubble)
            
    def think(self, text: str):
        self.say(text, Bubble.THINK)

    def sayforsecs(self, text: str, secs: int):
        self.say(text)
        self.sprite.code_manager.current_block.saying = True
        self.sprite.code_manager.current_block.add_to_wait_time = secs

    def thinkforsecs(self, text: str, secs: int):
        self.think(text)
        self.sprite.code_manager.current_block.saying = True
        self.sprite.code_manager.current_block.add_to_wait_time = secs

    def kill(self):
        if self.bubble:
            self.bubble.kill()
            self.bubble = None



import pygame
from pygame.mixer import music
from pystage.core.assets import SoundManager
from pystage.core._base_sprite import BaseSprite
import time


class _Sound(BaseSprite):

    # Like for costumes and backdrops, we need a class structure here.
    # Plus a global sound manager.
    def __init__(self):
        super().__init__()
        self.sound_manager = SoundManager(self)
        self.mixer = pygame.mixer
        self.mixer.init(channels=2)
        self.current_pan = 0
        self.current_pitch = 0
        self.current_volume = 100


    def pystage_addsound(self, name):
        self.sound_manager.add_sound(name)


    def sound_play(self, name, loop=0):
        channel = self.mixer.find_channel()
        sound = self.sound_manager.get_sound(name)
        if sound is not None:
            channel.play(sound, loop)
        return channel

    def sound_playuntildone(self, name):
        sound = self.sound_manager.get_sound(name)
        if sound is not None:
            self.mixer.find_channel().play(sound, 0)
            # time.sleep(sound.get_length())
            # This need to be done via wait time in code block
            # TODO: Add this function to yield blocks.
            self.code_manager.current_block.add_to_wait_time = sound.get_length()

    def sound_stopallsounds(self):
        self.mixer.stop()

    def sound_changeeffectby_pitch(self, value):
        # TODO: for pitching there is no ready to use code in pygame. To do so
        # we must operate on the audio array itself.
        # -360 to 360, 10 is a half-step, 120 an octave
        # changes only the speed of the sound
        pass

    sound_changeeffectby_pitch.opcode = "sound_changeeffectby"
    sound_changeeffectby_pitch.param = "EFFECT"
    sound_changeeffectby_pitch.value = "PITCH"
    sound_changeeffectby_pitch.translation = "sound_effects_pitch"

    def sound_changeeffectby_pan(self, value):
        # norm pan value from -100/100 to range 0/1
        self.current_pan += value
        self.current_pan = min(100, max(-100, self.current_pan))
        self._apply()

    sound_changeeffectby_pan.opcode = "sound_changeeffectby"
    sound_changeeffectby_pan.param = "EFFECT"
    sound_changeeffectby_pan.value = "PAN"
    sound_changeeffectby_pan.translation = "sound_effects_pan"

    def sound_seteffectto_pitch(self, value):
        # TODO: for pitching there is no ready to use code in pygame. To do so
        # we must operate on the audio array itself.
        pass

    sound_seteffectto_pitch.opcode = "sound_seteffectto"
    sound_seteffectto_pitch.param = "EFFECT"
    sound_seteffectto_pitch.value = "PITCH"
    sound_seteffectto_pitch.translation = "sound_effects_pitch"

    def sound_seteffectto_pan(self, value):
        # Values from -100 (left) to 100 (right)
        self.current_pan = value
        self.current_pan = min(100, max(-100, self.current_pan))
        self._apply()

    sound_seteffectto_pan.opcode = "sound_seteffectto"
    sound_seteffectto_pan.param = "EFFECT"
    sound_seteffectto_pan.value = "PAN"
    sound_seteffectto_pan.translation = "sound_effects_pan"

    def sound_cleareffects(self):
        self.current_pan = 0
        self.current_pitch = 0
        self._apply()

        # apply pitch

    def _apply(self):
        # norm pan value from -100/100 to range 0/1
        pgpan = (self.current_pan + 100) / 200
        pgvolume = self.current_volume / 100
        for channel_id in range(self.mixer.get_num_channels()):
            if pgpan > 0.5:
                self.mixer.Channel(channel_id).set_volume(1, 0)
            else:
                self.mixer.Channel(channel_id).set_volume(0, 1)

        for channel_id in range(self.mixer.get_num_channels()):
            self.mixer.Channel(channel_id).set_volume(pgvolume)

    def sound_changevolumeby(self, value):
        self.current_volume += value
        self.current_volume = min(100, max(0, self.current_volume))
        self._apply()

    def sound_setvolumeto(self, value):
        self.current_volume = value
        self.current_volume = min(100, max(0, self.current_volume))
        self._apply()

    def sound_volume(self):
        # as we hide the channel mechanic, we assume all channels are set to the same volume
        return self.mixer.Channel(0).get_volume() * 100




from pystage.en import Sprite, Stage

stage = Stage()
zombie = stage.add_a_sprite()


def doit(zombie: Sprite):
    zombie.set_rotation_style_left_right()
    while True:
        zombie.move(1)
        zombie.if_on_edge_bounce()

def gotomouse(zombie: Sprite):
    zombie.point_towards_mouse_pointer()
    print(zombie.direction())

zombie.when_program_starts(doit)
zombie.when_key_pressed(" ", gotomouse)

stage.play()



import argparse
import json
import logging
import math
import os
import re
import sys
import textwrap
import zipfile
from collections import OrderedDict
from pathlib import Path
import importlib
from pystage.l10n.api import get_translated_function

from pystage.convert import CodeWriter, sb3_templates

logger = logging.getLogger(__name__)

'''
For the conversion, we create a simplified version of the
block trees that is used in the code template functions.
'''


# Target (intermediate) structure
# project = {
#         "costumes": {
#           "abcd...": {
#               "global_name": "cat1",
#               "extension": "png",
#               },
#           }
#         "sounds": {
#           "abcd...": {
#               "global_name": "meeow",
#               "extension": "wav",
#               },
#           }
#         "stage": {
#               "name": "Stage",
#               "blocks": [],
#               "costumes": [],
#               "currentCostume": 0
#           }
#         "sprites": [
#             {
#                 "name": "Sprite1",
#                 "blocks": [
#                     { # BLOCK
#                         "opcode": "event_whenthisspriteclicked",
#                         "stage": False,
#                         "params": {
#                             "BROADCAST_INPUT": {"start",
#                             "COSTUME": "BLOCK"
#                             }
#                         "next": "BLOCK",
#                         "comments": ["text"],
#                         }
#                     ]
#                 "costumes": [{
#                       "md5": "abcd..."
#                       "local_name": "cat1",
#                       "rotationCenterX": 240,
#                       "rotationCenterY": 180,
#                       },
#                   ],
#                 "sounds": [{
#                       "md5": "abcd..."
#                       "local_name": "cat1",
#                       },
#                   ],
#                 "currentCostume": 0
#                 },
#
#             ],
#
#         }

class DictClass(OrderedDict):
    '''
    A dictionary that can be accessed via dot notation for convenience.
    '''

    def __init__(self):
        super().__init__()

    def __getattr__(self, item):
        return self[item]


def get_input_value(i, stage):
    '''
    Extracts values from sb3 input representations.
    '''
    # 4: number, 5: positive number, 6: positive integer
    # 7: integer, 8: angle, 9: color, 10: string
    # 11: broadcast, 12: variable, 13: list

    input_type = i[0]
    value = i[1]

    if input_type in [4, 5, 8]:
        if isinstance(value, str) and value.strip() == '':
            value = 0
        else:
            value = float(value)
    elif input_type in [6, 7]:
        if isinstance(value, str) and value.strip() == '':
            value = 0
        else:
            value = int(value)
    elif input_type == 9:
        h = value.lstrip("#")
        value = tuple(int(h[i:i + 2], 16) for i in (0, 2, 4))
    elif input_type == 12:
        block = DictClass()
        block.update({
                "opcode": "data_variable",
                "params": {
                    "VARIABLE": f'"{value}"',
                    },
                "next": False,
                "stage": stage,
                })
        return block
    elif input_type in [10, 11, 13]:
        try:
            f = float(value)
            if math.floor(f) == f:
                f = int(f)
            value = f
        except:
            value = f'"{value}"'
    return value


def get_block(block, blocks, stage):
    '''
    Recursively turns the sb3 structure into our intermediate representation.
    '''
    res = DictClass()
    res.update({
        "opcode": block["opcode"],
        "params": DictClass(),
        "next": False,
        "stage": stage
    })
    for f in block["fields"]:
        res["params"][f] = f'"{block["fields"][f][0]}"'
    for i in block["inputs"]:
        value = block["inputs"][i][1]
        if not value:
            # empty if condition
            continue
        if isinstance(value, list):
            res["params"][i] = get_input_value(value, stage)
        else:
            res["params"][i] = get_block(blocks[value], blocks, stage)
    if block["next"]:
        res["next"] = get_block(blocks[block["next"]], blocks, stage)
    return res


def to_filename(name: str):
    umlauts = {"ö": "oe", "ü": "ue", "ä": "ae", "ß": "ss"}
    name = name.lower().strip()
    for c in umlauts:
        name = name.replace(c, umlauts[c])
    name = re.sub(r"[^a-z0-9_]", "_", name)
    name = re.sub(r"_+", "_", name)
    return name

# Used to check for duplicate names
global_names = set()

def unique_global_name(name, ext):
    iname = name
    i = 2
    while f"{iname}.{ext}" in global_names:
        iname = f"{name}_{i}"
        i += 1
    global_names.add(f"{iname}.{ext}")
    return iname, ext

def get_sb3_data(filename):
    archive = zipfile.ZipFile(filename, 'r')
    with archive.open("project.json") as f:
        data = json.loads(f.read())
    return data

def get_intermediate(data, name):
    hat_blocks = [
        "event_whenthisspriteclicked",
        "event_whenbroadcastreceived",
        "event_whenflagclicked",
        "event_whenkeypressed",
        "event_whenbackdropswitchesto",
        "event_whengreaterthan",
        "control_start_as_clone",
    ]

    project = DictClass()
    project.update({
        "name": name,
        "sprites": [],
        "costumes": {},
        "sounds": {},
    })


    for target in data["targets"]:
        sprite = DictClass()
        sprite.update({
            "name": target['name'],
            "blocks": [],
            "costumes": [],
            "sounds": [],
            "variables": {}, # name: value
            "monitors": [], # name: value
            "currentCostume": target["currentCostume"],
            "layerOrder": target["layerOrder"],
            "visible": target["visible"] if "visible" in target else True,
            "x": target["x"] if "x" in target else 0,
            "y": target["y"] if "y" in target else 0,
            "size": target["size"] if "size" in target else 100,
            "volume": target["volume"] if "volume" in target else 100,
            "direction": target["direction"] if "direction" in target else 90,
            "rotationStyle": target["rotationStyle"] if "rotationStyle" in target else "all around",
        })
        if target["isStage"]:
            project["stage"] = sprite
        else:
            project["sprites"].append(sprite)
        sprite["currentCostume"] = target["currentCostume"]
        blocks = target["blocks"]
        for key in blocks:
            b = blocks[key]
            if isinstance(b, list):
                # Stuff like variables, not interesting at this point
                continue
            # if b["parent"] is None:
            if b["opcode"] in hat_blocks:
                block = get_block(b, blocks, target["isStage"])
                sprite["blocks"].append(block)
        for c in target["costumes"]:
            if c["assetId"] not in project["costumes"]:
                name, ext = unique_global_name(to_filename(c["name"]), c["dataFormat"])
                project["costumes"][c["assetId"]] = {
                    "global_name": name,
                    "extension": ext,
                }
            sprite["costumes"].append({
                "md5": c["assetId"],
                "local_name": c["name"],
                "bitmapResolution": c["bitmapResolution"] if "bitmapResolution" in c else 1,
                "rotationCenterX": c["rotationCenterX"],
                "rotationCenterY": c["rotationCenterY"],
            })
        for s in target["sounds"]:
            if s["assetId"] not in project["sounds"]:
                name, ext = unique_global_name(to_filename(s["name"]), s["dataFormat"])
                project["sounds"][s["assetId"]] = {
                    "global_name": name,
                    "extension": ext,
                }
            sprite["sounds"].append({
                "md5": s["assetId"],
                "local_name": s["name"],
            })
        for key in target["variables"]:
            v = target["variables"][key]
            variable = {}
            sprite["variables"][v[0]] = v[1]

    for m in data["monitors"]:
        if not m["visible"]:
            continue
        monitor = {}
        if not m["spriteName"]:
            sprite = project["stage"]
        else:
            for s in project["sprites"]:
                if s["name"] == m["spriteName"]:
                    sprite = sprite
                    break
        monitor["style"] = m["mode"]
        if monitor["style"] == "default":
            monitor["style"] == "normal"
        monitor["x"] = m["x"]
        monitor["y"] = m["y"]
        monitor["opcode"] = m["opcode"]
        if "VARIABLE" in m["params"]:
            monitor["variable"] = m["params"]["VARIABLE"]
        sprite["monitors"].append(monitor)

    return project


def get_python(project, language="core", project_link=None):
    lang_module = importlib.import_module(f"pystage.{language}")
    sprite_class = lang_module.sprite_class.__name__
    stage_class = lang_module.stage_class.__name__
    add_backdrop = get_translated_function("pystage_addbackdrop", language, stage=True)
    add_costume = get_translated_function("pystage_addcostume", language)
    add_sound = get_translated_function("pystage_addsound", language)
    add_variable = get_translated_function("pystage_makevariable", language)
    create_sprite = get_translated_function("pystage_createsprite", language, stage=True)
    play = get_translated_function("pystage_play", language, stage=True)
    res = ""
    if project_link:
        res += textwrap.dedent(f'''\
        # Online Converter, URL: {project_link}
        ''')
    res += textwrap.dedent(f'''\
            # {project['name']} (pyStage, converted from Scratch 3)

            from pystage.{language} import {sprite_class}, {stage_class}

            ''')
    writer = CodeWriter(project, sb3_templates.templates, language)
    writer.set_sprite(project["stage"]["name"])
    stage_var = writer.get_sprite_var()
    backdrops = []
    for bd in project["stage"]["costumes"]:
        backdrops.append(writer.global_backdrop(bd["local_name"], False))
    res += textwrap.dedent(f'''\
            {stage_var} = {stage_class}()
            ''')
    for bd in backdrops:
        res += textwrap.dedent(f'''\
                {stage_var}.{add_backdrop}('{bd}')
                ''')
    for v in project["stage"]["variables"]:
        res += textwrap.dedent(f'''\
                {stage_var}.{add_variable}('{v}')
                ''')
    for monitor in project["stage"]["monitors"]:
        # Only variable monitors are currently implemented
        if not "variable" in monitor:
            continue
        res += textwrap.dedent(f'''\
                {stage_var}.{get_translated_function("data_showvariable", language)}("{monitor["variable"]}")
                {stage_var}.{get_translated_function("pystage_setmonitorposition", language)}("{monitor["variable"]}", {-240 + monitor["x"]}, {180 - monitor["y"]})
                ''')
        if monitor["style"] == "large":
            res += textwrap.dedent(f'''\
                    {stage_var}.{get_translated_function("pystage_setmonitorstyle_large", language)}("{monitor["variable"]}")
                    ''')
    for block in project["stage"]["blocks"]:
        res += writer.process(block)
    for sprite in project["sprites"]:
        writer.set_sprite(sprite["name"])
        sprite_var = writer.get_sprite_var()
        costumes = [(writer.global_costume(c["local_name"], False), c) for c in sprite["costumes"]]
        sounds = [writer.global_sound(s["local_name"], False) for s in sprite["sounds"]]
        
        # Generate initialization code for each sprite.
        res += textwrap.dedent(f"""\
                
                # Create and initialize sprite '{sprite_var}'
                """)
        res += textwrap.dedent(f'''\
                {sprite_var} = {stage_var}.{create_sprite}(None)
                {sprite_var}.{get_translated_function("pystage_setname", language)}("{sprite["name"]}")
                {sprite_var}.{get_translated_function("motion_setx", language)}({sprite["x"]})
                {sprite_var}.{get_translated_function("motion_sety", language)}({sprite["y"]})
                {sprite_var}.{get_translated_function("looks_gotofrontback_back", language)}()
                {sprite_var}.{get_translated_function("looks_goforwardbackwardlayers_forward", language)}({sprite["layerOrder"]})
                ''')
        if sprite["direction"] != 90:
            res += textwrap.dedent(f"""\
                    {sprite_var}.{get_translated_function("motion_pointindirection", language)}({sprite["direction"]})
                    """)
        if sprite["size"] != 100:
            res += textwrap.dedent(f"""\
                    {sprite_var}.{get_translated_function("looks_setsizeto", language)}({sprite["size"]})
                    """)
        if not sprite["visible"]:
            res += textwrap.dedent(f"""\
                    {sprite_var}.{get_translated_function("looks_hide", language)}()
                    """)
        if sprite["volume"] != 100:
            res += textwrap.dedent(f"""\
                    {sprite_var}.{get_translated_function("sound_setvolumeto", language)}({sprite["volume"]})
                    """)
        if sprite["rotationStyle"] != "all around":
            print("WARNING: preset rotation styles not yet implemented!")
        for c in costumes:
            factor = ""
            if c[1]["bitmapResolution"] != 1:
                factor=f", factor={c[1]['bitmapResolution']}"
            res += textwrap.dedent(f'''\
                {sprite_var}.{add_costume}('{c[0]}', center_x={c[1]["rotationCenterX"]}, center_y={c[1]["rotationCenterY"]}{factor})
                ''')
        if sprite["currentCostume"] != 0:
            for i in range(sprite["currentCostume"]):
                res += textwrap.dedent(f'''\
                    {sprite_var}.{get_translated_function("looks_nextcostume", language)}()
                    ''')

        for s in sounds:
            res += textwrap.dedent(f'''\
                {sprite_var}.{add_sound}('{s}')
                ''')

        for v in sprite["variables"]:
            res += textwrap.dedent(f'''\
                    {sprite_var}.{add_variable}('{v}')
                    ''')
        for monitor in sprite["monitors"]:
            # Only variable monitors are currently implemented
            if not "variable" in monitor:
                continue
            res += textwrap.dedent(f'''\
                    {sprite_var}.{get_translated_function("data_showvariable", language)}("{monitor["variable"]}")
                    {sprite_var}.{get_translated_function("pystage_setmonitorposition", language)}("{monitor["variable"]}", {-240 + monitor["x"]}, {180 - monitor["y"]})
                    ''')
            if monitor["style"] == "large":
                res += textwrap.dedent(f'''\
                        {sprite_var}.{get_translated_function("pystage_setmonitorstyle_large", language)}("{monitor["variable"]}")
                        ''')
    
    # Now generate the code for each sprite.
    for sprite in project["sprites"]:
        writer.set_sprite(sprite["name"])
        sprite_var = writer.get_sprite_var()
        res += textwrap.dedent(f"""\
                
                # Scratch Blocks for '{sprite_var}'
                """)
        for block in sprite["blocks"]:
            res += writer.process(block)


    res += textwrap.dedent(f'''\

            {stage_var}.{play}()
            ''')
    return res


def print_python(project, language="core"):
    print(get_python(project, language))


def create_project(file_name, project_name, project, directory, language="core", project_link=None):
    print(f"Creating project: {project_name}")
    print(f"Exporting to: {directory}")
    archive = zipfile.ZipFile(file_name, 'r')
    dp = Path(directory)
    if dp.exists():
        raise("Directory must not exist at this point, but it does: {}".format(directory))
    os.mkdir(dp)
    os.mkdir(dp / "images")
    os.mkdir(dp / "sounds")
    for key in project["costumes"]:
        c = project["costumes"][key]
        with archive.open(f'{key}.{c["extension"]}') as infile:
            with open(dp / "images" / f'{c["global_name"]}.{c["extension"]}', "wb") as outfile:
                outfile.write(infile.read())
    for key in project["sounds"]:
        s = project["sounds"][key]
        with archive.open(f'{key}.{s["extension"]}') as infile:
            with open(dp / "sounds" / f'{s["global_name"]}.{s["extension"]}', "wb") as outfile:
                outfile.write(infile.read())
    with open(dp / f'{project_name}.py', "w", encoding="utf-8") as pyfile:
        pyfile.write(get_python(project, language=language, project_link=project_link))



##
# Main
#


if __name__ == "__main__":

    parser = argparse.ArgumentParser(
        prog="python -m pystage.convert.sb3",
        description="Convert Scratch 3 (sb3) files to Python.")
    parser.add_argument("file", metavar="FILE", type=str, help="the sb3 file to be converted")
    parser.add_argument("-i", "--intermediate", action="store_true", help="print intermediate code representation")
    parser.add_argument("-s", "--sb3-json", action="store_true", help="print sb3 project.json")
    parser.add_argument("-p", "--python", action="store_true", help="print python code")
    parser.add_argument("-l", "--language", metavar="LANG", type=str, help="API language, 2-letter ISO code, e.g. en, de, ...", nargs="?", default="core")
    parser.add_argument("-v", "--verbose", action="store_true", help="be verbose (info)")
    parser.add_argument("-vv", "--debug", action="store_true", help="print debug information")
    parser.add_argument("-d", "--directory", metavar="DIR", type=str, help="the project directory")
    args = parser.parse_args()

    if args.debug:
        logging.basicConfig(level=logging.DEBUG)
    elif args.verbose:
        logging.basicConfig(level=logging.INFO)

    project_name = to_filename(Path(args.file).stem)
    data = get_sb3_data(args.file)
    if args.sb3_json:
        print(json.dumps(data, indent=2))
        sys.exit(0)
    project = get_intermediate(data, project_name)
    if args.intermediate:
        print(json.dumps(project, indent=2))
        sys.exit(1)
    elif args.python:
        print_python(project, args.language)
        sys.exit(1)
    else:
        directory = project_name
        if args.directory:
            directory = args.directory
        print(f"Exporting to: {directory}")
        dp = Path(directory)
        if dp.exists() and not dp.is_dir:
            print("Output directory exists, but is not a directory. Use -d to specify another directory.")
            sys.exit(1)
        elif dp.exists() and next(dp.iterdir(), False):
            print("Output directory is not empty. Use -d to specify another one.")
            sys.exit(1)
        elif not dp.exists():
            create_project(args.file, project_name, project, directory, args.language)



from pystage.de import Bühne, Figur

bühne = Bühne()
zombie = bühne.füge_eine_figur_hinzu()

def mach_was(zombie : Figur):
    zombie.gehe_zu_x_y(0, 0)
    for i in range(4):
        zombie.gehe_er_schritt(20)
        zombie.warte_sekunden(1)
        zombie.drehe_dich_nach_links_um_grad(90)
        zombie.warte_sekunden(1)

zombie.wenn_GREENFLAG_angeklickt_wird(mach_was)


def reagiere(zombie):
    zombie.gehe_er_schritt(10)

zombie.wenn_taste_gedrückt_wird(" ", reagiere)

bühne.abspielen()




from pystage.core.sprite import CoreSprite


class Figur():
    def __init__(self, core_sprite):
        self._core : CoreSprite = core_sprite
        self._core.facade = self


            
    def erzeuge_klon_von_mir(self):
        """erzeuge Klon von mir selbst

        Translation string: erzeuge Klon von %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Returns
        -------

        """
        return self._core.control_create_clone_of_myself()
                
    def erzeuge_klon_von_sprite(self, sprite):
        """erzeuge Klon von %1

        Translation string: erzeuge Klon von %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.control_create_clone_of_sprite(sprite)

    def lösche_diesen_klon(self):
        """lösche diesen Klon

        Translation string: lösche diesen Klon
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.control_delete_this_clone()
                
    def wenn_ich_als_klon_entstehe(self, generator_function, name='', no_refresh=False):
        """Wenn ich als Klon entstehe

        Translation string: Wenn ich als Klon entstehe
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.control_start_as_clone(generator_function, name, no_refresh)
                
    def stoppe_alles(self):
        """stoppe alles

        Translation string: stoppe alles
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.control_stop_all()
                
    def stoppe_andere_skripte_der_figur(self):
        """stoppe andere Skripte der Figur

        Translation string: stoppe andere Skripte der Figur
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.control_stop_other()
                
    def stoppe_dieses_skript(self):
        """stoppe dieses Skript

        Translation string: stoppe dieses Skript
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.control_stop_this()
                
    def warte_sekunden(self, secs):
        """warte %1 Sekunden

        Translation string: warte %1 Sekunden
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        secs : FILL
        

        Returns
        -------

        """
        return self._core.control_wait(secs)
                
    def ändere_um(self, name, value):
        """ändere %1 um %2

        Translation string: ändere %1 um %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        value : FILL
        

        Returns
        -------

        """
        return self._core.data_changevariableby(name, value)
                
    def verstecke_variable(self, name):
        """verstecke Variable %1

        Translation string: verstecke Variable %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.data_hidevariable(name)
                
    def setze_auf(self, name, value):
        """setze %1 auf %2

        Translation string: setze %1 auf %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        value : FILL
        

        Returns
        -------

        """
        return self._core.data_setvariableto(name, value)
                
    def zeige_variable(self, name):
        """zeige Variable %1

        Translation string: zeige Variable %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.data_showvariable(name)
                
    def data_variable(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.data_variable(name)
                
    def sende_an_alle(self, message):
        """sende %1 an alle

        Translation string: sende %1 an alle
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        message : FILL
        

        Returns
        -------

        """
        return self._core.event_broadcast(message)
                
    def sende_an_alle_und_warte(self, message):
        """sende %1 an alle und warte

        Translation string: sende %1 an alle und warte
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        message : FILL
        

        Returns
        -------

        """
        return self._core.event_broadcastandwait(message)
                
    def wenn_das_bühnenbild_zu_wechselt(self, backdrop, generator_function, name='', no_refresh=False):
        """Wenn das Bühnenbild zu %1 wechselt

        Translation string: Wenn das Bühnenbild zu %1 wechselt
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        backdrop : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenbackdropswitchesto(backdrop, generator_function, name, no_refresh)
                
    def wenn_ich_empfange(self, message, generator_function, name='', no_refresh=False):
        """Wenn ich %1 empfange

        Translation string: Wenn ich %1 empfange
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        message : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenbroadcastreceived(message, generator_function, name, no_refresh)
                
    def wenn_GREENFLAG_angeklickt_wird(self, generator_function, name='', no_refresh=False):
        """Wenn <greenflag> angeklickt wird

        Translation string: Wenn <greenflag> angeklickt wird
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenflagclicked(generator_function, name, no_refresh)
                
    def wenn_lautstärke_GREATERTHAN(self, value, generator_function, name='', no_refresh=False):
        """Wenn Lautstärke <greater> %2

        Translation string: Wenn Lautstärke <greater> %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whengreaterthan_loudness(value, generator_function, name, no_refresh)
                
    def wenn_stoppuhr_GREATERTHAN(self, value, generator_function, name='', no_refresh=False):
        """Wenn Stoppuhr <greater> %2

        Translation string: Wenn Stoppuhr <greater> %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whengreaterthan_timer(value, generator_function, name, no_refresh)
                
    def wenn_taste_gedrückt_wird(self, key, generator_function, name='', no_refresh=False):
        """Wenn Taste %1 gedrückt wird

        Translation string: Wenn Taste %1 gedrückt wird
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        key : FILL
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenkeypressed(key, generator_function, name, no_refresh)
                
    def wenn_diese_figur_angeklickt_wird(self, generator_function, name='', no_refresh=False):
        """Wenn diese Figur angeklickt wird

        Translation string: Wenn diese Figur angeklickt wird
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        generator_function : FILL
        name : FILL
        no_refresh : FILL
        

        Returns
        -------

        """
        return self._core.event_whenthisspriteclicked(generator_function, name, no_refresh)
                
    def bühnenbild_name(self):
        """Bühnenbild Name

        Translation string: Bühnenbild Name
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_backdropnumbername_name()
                
    def bühnenbild_nummer(self):
        """Bühnenbild Nummer

        Translation string: Bühnenbild Nummer
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_backdropnumbername_number()
                
    def ändere_effekt_helligkeit_um(self, value):
        """ändere Effekt Helligkeit um %2

        Translation string: ändere Effekt Helligkeit um %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_brightness(value)
                
    def ändere_effekt_farbe_um(self, value):
        """ändere Effekt Farbe um %2

        Translation string: ändere Effekt Farbe um %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_color(value)
                
    def ändere_effekt_fischauge_um(self, value):
        """ändere Effekt Fischauge um %2

        Translation string: ändere Effekt Fischauge um %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_fisheye(value)
                
    def ändere_effekt_durchsichtigkeit_um(self, value):
        """ändere Effekt Durchsichtigkeit um %2

        Translation string: ändere Effekt Durchsichtigkeit um %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_ghost(value)
                
    def ändere_effekt_mosaik_um(self, value):
        """ändere Effekt Mosaik um %2

        Translation string: ändere Effekt Mosaik um %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_mosaic(value)
                
    def ändere_effekt_pixel_um(self, value):
        """ändere Effekt Pixel um %2

        Translation string: ändere Effekt Pixel um %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_pixelate(value)
                
    def ändere_effekt_wirbel_um(self, value):
        """ändere Effekt Wirbel um %2

        Translation string: ändere Effekt Wirbel um %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_changeeffectby_whirl(value)
                
    def ändere_größe_um(self, percent):
        """ändere Größe um %1

        Translation string: ändere Größe um %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        percent : FILL
        

        Returns
        -------

        """
        return self._core.looks_changesizeby(percent)
                
    def schalte_grafikeffekte_aus(self):
        """schalte Grafikeffekte aus

        Translation string: schalte Grafikeffekte aus
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_cleargraphiceffects()
                
    def kostüm_name(self):
        """Kostüm Name

        Translation string: Kostüm Name
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_costumenumbername_name()
                
    def kostüm_nummer(self):
        """Kostüm Nummer

        Translation string: Kostüm Nummer
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_costumenumbername_number()
                
    def gehe_ebenen_nach_hinten(self, value):
        """gehe %2 Ebenen nach hinten

        Translation string: gehe %2 Ebenen nach hinten
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_goforwardbackwardlayers_backward(value)
                
    def gehe_ebenen_nach_vorne(self, value):
        """gehe %2 Ebenen nach vorne

        Translation string: gehe %2 Ebenen nach vorne
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_goforwardbackwardlayers_forward(value)
                
    def gehe_zu_hinterster_ebene(self):
        """gehe zu hinterster Ebene

        Translation string: gehe zu hinterster Ebene
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_gotofrontback_back()
                
    def gehe_zu_vorderster_ebene(self):
        """gehe zu vorderster Ebene

        Translation string: gehe zu vorderster Ebene
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_gotofrontback_front()
                
    def verstecke_dich(self):
        """verstecke dich

        Translation string: verstecke dich
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_hide()
                
    def nächstes_bühnenbild(self):
        """nächstes Bühnenbild

        Translation string: nächstes Bühnenbild
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_nextbackdrop()
                
    def wechsle_zum_nächsten_kostüm(self):
        """wechsle zum nächsten Kostüm

        Translation string: wechsle zum nächsten Kostüm
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_nextcostume()
                
    def sage(self, text):
        """sage %1

        Translation string: sage %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        text : FILL
        

        Returns
        -------

        """
        return self._core.looks_say(text)
                
    def sage_für_sekunden(self, text, secs):
        """sage %1 für %2 Sekunden

        Translation string: sage %1 für %2 Sekunden
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        text : FILL
        secs : FILL
        

        Returns
        -------

        """
        return self._core.looks_sayforsecs(text, secs)
                
    def setze_effekt_helligkeit_auf(self, value):
        """setze Effekt Helligkeit auf %2

        Translation string: setze Effekt Helligkeit auf %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_brightness(value)
                
    def setze_effekt_farbe_auf(self, value):
        """setze Effekt Farbe auf %2

        Translation string: setze Effekt Farbe auf %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_color(value)
                
    def setze_effekt_fischauge_auf(self, value):
        """setze Effekt Fischauge auf %2

        Translation string: setze Effekt Fischauge auf %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_fisheye(value)
                
    def setze_effekt_durchsichtigkeit_auf(self, value):
        """setze Effekt Durchsichtigkeit auf %2

        Translation string: setze Effekt Durchsichtigkeit auf %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_ghost(value)
                
    def setze_effekt_mosaik_auf(self, value):
        """setze Effekt Mosaik auf %2

        Translation string: setze Effekt Mosaik auf %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_mosaic(value)
                
    def setze_effekt_pixel_auf(self, value):
        """setze Effekt Pixel auf %2

        Translation string: setze Effekt Pixel auf %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_pixelate(value)
                
    def setze_effekt_wirbel_auf(self, value):
        """setze Effekt Wirbel auf %2

        Translation string: setze Effekt Wirbel auf %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.looks_seteffectto_whirl(value)
                
    def setze_größe_auf(self, percent):
        """setze Größe auf %1

        Translation string: setze Größe auf %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        percent : FILL
        

        Returns
        -------

        """
        return self._core.looks_setsizeto(percent)
                
    def zeige_dich(self):
        """zeige dich

        Translation string: zeige dich
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_show()
                
    def größe(self):
        """Größe

        Translation string: Größe
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.looks_size()
                
    def wechsle_zu_bühnenbild(self, backdrop):
        """wechsle zu Bühnenbild %1

        Translation string: wechsle zu Bühnenbild %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        backdrop : FILL
        

        Returns
        -------

        """
        return self._core.looks_switchbackdropto(backdrop)
                
    def wechsle_zu_kostüm(self, costume):
        """wechsle zu Kostüm %1

        Translation string: wechsle zu Kostüm %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        costume : FILL
        

        Returns
        -------

        """
        return self._core.looks_switchcostumeto(costume)
                
    def denke(self, text):
        """denke %1

        Translation string: denke %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        text : FILL
        

        Returns
        -------

        """
        return self._core.looks_think(text)
                
    def denke_für_sekunden(self, text, secs):
        """denke %1 für %2 Sekunden

        Translation string: denke %1 für %2 Sekunden
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        text : FILL
        secs : FILL
        

        Returns
        -------

        """
        return self._core.looks_thinkforsecs(text, secs)
                
    def ändere_x_um(self, value):
        """ändere x um %1

        Translation string: ändere x um %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.motion_changexby(value)
                
    def ändere_y_um(self, value):
        """ändere y um %1

        Translation string: ändere y um %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.motion_changeyby(value)
                
    def richtung(self):
        """Richtung

        Translation string: Richtung
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_direction()
                
    def gleite_in_sek_zu_x_y(self, secs, x, y):
        """gleite in %1 Sek. zu x:%2  y:%3

        Translation string: gleite in %1 Sek. zu x:%2  y:%3
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        secs : FILL
        x : FILL
        y : FILL
        

        Returns
        -------

        """
        return self._core.motion_glidesecstoxy(secs, x, y)
                
    def gleite_in_sek_zu_mauszeiger(self, secs):
        """gleite in %1 Sek. zu Mauszeiger

        Translation string: gleite in %1 Sek. zu Mauszeiger
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        secs : FILL
        

        Returns
        -------

        """
        return self._core.motion_glideto_pointer(secs)
                
    def gleite_in_sek_zu_zufallsposition(self, secs):
        """gleite in %1 Sek. zu Zufallsposition

        Translation string: gleite in %1 Sek. zu Zufallsposition
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        secs : FILL
        

        Returns
        -------

        """
        return self._core.motion_glideto_random(secs)
                
    def gleite_in_sek_zu(self, secs, sprite):
        """gleite in %1 Sek. zu %2

        Translation string: gleite in %1 Sek. zu %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        secs : FILL
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.motion_glideto_sprite(secs, sprite)
                
    def gehe_zu_mauszeiger(self):
        """gehe zu Mauszeiger

        Translation string: gehe zu Mauszeiger
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_goto_pointer()
                
    def gehe_zu_zufallsposition(self):
        """gehe zu Zufallsposition

        Translation string: gehe zu Zufallsposition
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_goto_random()
                
    def gehe_zu(self, sprite):
        """gehe zu %1

        Translation string: gehe zu %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.motion_goto_sprite(sprite)
                
    def gehe_zu_x_y(self, x, y):
        """gehe zu x: %1 y: %2

        Translation string: gehe zu x: %1 y: %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        x : FILL
        y : FILL
        

        Returns
        -------

        """
        return self._core.motion_gotoxy(x, y)
                
    def pralle_vom_rand_ab(self):
        """pralle vom Rand ab

        Translation string: pralle vom Rand ab
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_ifonedgebounce()
                
    def gehe_er_schritt(self, steps):
        """gehe %1 er Schritt

        Translation string: gehe %1 er Schritt
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        steps : FILL
        

        Returns
        -------

        """
        return self._core.motion_movesteps(steps)
                
    def setze_richtung_auf_grad(self, direction):
        """setze Richtung auf %1 Grad

        Translation string: setze Richtung auf %1 Grad
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        direction : FILL
        

        Returns
        -------

        """
        return self._core.motion_pointindirection(direction)
                
    def drehe_dich_zu_mauszeiger(self):
        """drehe dich zu Mauszeiger

        Translation string: drehe dich zu Mauszeiger
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_pointtowards_pointer()
                
    def drehe_dich_zu(self, sprite):
        """drehe dich zu %1

        Translation string: drehe dich zu %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.motion_pointtowards_sprite(sprite)
                
    def setze_drehtyp_auf_rundherum(self):
        """setze Drehtyp auf rundherum

        Translation string: setze Drehtyp auf rundherum
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_setrotationstyle_allaround()
                
    def setze_drehtyp_auf_nicht_drehen(self):
        """setze Drehtyp auf nicht drehen

        Translation string: setze Drehtyp auf nicht drehen
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_setrotationstyle_dontrotate()
                
    def setze_drehtyp_auf_helligkeit(self):
        """setze Drehtyp auf Helligkeit

        Translation string: setze Drehtyp auf Helligkeit
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_setrotationstyle_leftright()
                
    def setze_x_auf(self, value):
        """setze x auf %1

        Translation string: setze x auf %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.motion_setx(value)
                
    def setze_y_auf(self, value):
        """setze y auf %1

        Translation string: setze y auf %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.motion_sety(value)
                
    def drehe_dich_nach_links_um_grad(self, deg):
        """drehe dich nach links um %2 Grad

        Translation string: drehe dich nach links um %2 Grad
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        deg : FILL
        

        Returns
        -------

        """
        return self._core.motion_turnleft(deg)
                
    def drehe_dich_nach_rechts_um_grad(self, deg):
        """drehe dich nach rechts um %2 Grad

        Translation string: drehe dich nach rechts um %2 Grad
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        deg : FILL
        

        Returns
        -------

        """
        return self._core.motion_turnright(deg)
                
    def x_position(self):
        """x-Position

        Translation string: x-Position
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_xposition()
                
    def y_position(self):
        """y-Position

        Translation string: y-Position
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.motion_yposition()
                
    def von(self, operator, number):
        """%1 von %2

        Translation string: %1 von %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        operator : FILL
        number : FILL
        

        Returns
        -------

        """
        return self._core.operator_mathop(operator, number)
                
    def zufallszahl_von_bis(self, start, end):
        """Zufallszahl von %1 bis %2

        Translation string: Zufallszahl von %1 bis %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        start : FILL
        end : FILL
        

        Returns
        -------

        """
        return self._core.operator_random(start, end)
                
    def ändere_stift_helligkeit_um(self, value):
        """ändere Stift Helligkeit um [VALUE]

        Translation string: ändere Stift Helligkeit um [VALUE]
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_changePenColorParamBy_brightness(value)
                
    def ändere_stift_farbe_um(self, value):
        """ändere Stift Farbe um [VALUE]
        TODO TRANSLATORS: Needs to be distinguished from setPenColorToColor. This is only the hue value.
        

        Translation string: ändere Stift Farbe um [VALUE]
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_changePenColorParamBy_color(value)
                
    def ändere_stift_sättigung_um(self, value):
        """ändere Stift Sättigung um [VALUE]

        Translation string: ändere Stift Sättigung um [VALUE]
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_changePenColorParamBy_saturation(value)
                
    def ändere_stift_transparenz_um(self, value):
        """ändere Stift Transparenz um [VALUE]

        Translation string: ändere Stift Transparenz um [VALUE]
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_changePenColorParamBy_transparency(value)
                
    def ändere_stiftdicke_um(self, value):
        """ändere Stiftdicke um [SIZE]

        Translation string: ändere Stiftdicke um [SIZE]
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_changePenSizeBy(value)
                
    def lösche_alles(self):
        """lösche alles

        Translation string: lösche alles
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.pen_clear()
                
    def schalte_stift_ein(self):
        """schalte Stift ein

        Translation string: schalte Stift ein
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.pen_penDown()
                
    def schalte_stift_aus(self):
        """schalte Stift aus

        Translation string: schalte Stift aus
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.pen_penUp()
                
    def setze_stift_helligkeit_auf(self, value):
        """setze Stift Helligkeit auf [VALUE]

        Translation string: setze Stift Helligkeit auf [VALUE]
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_setPenColorParamTo_brightness(value)
                
    def setze_stift_farbe_auf(self, value):
        """setze Stift Farbe auf [VALUE]
        TODO TRANSLATORS: Needs to be distinguished from setPenColorToColor. This is only the hue value.
        

        Translation string: setze Stift Farbe auf [VALUE]
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_setPenColorParamTo_color(value)
                
    def setze_stift_sättigung_auf(self, value):
        """setze Stift Sättigung auf [VALUE]

        Translation string: setze Stift Sättigung auf [VALUE]
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_setPenColorParamTo_saturation(value)
                
    def setze_stift_transparenz_auf(self, value):
        """setze Stift Transparenz auf [VALUE]

        Translation string: setze Stift Transparenz auf [VALUE]
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_setPenColorParamTo_transparency(value)
                
    def setze_stiftfarbe_auf(self, color):
        """setze Stiftfarbe auf [COLOR]

        Translation string: setze Stiftfarbe auf [COLOR]
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        color : FILL
        

        Returns
        -------

        """
        return self._core.pen_setPenColorToColor(color)
                
    def setze_stiftdicke_auf(self, value):
        """setze Stiftdicke auf [SIZE]

        Translation string: setze Stiftdicke auf [SIZE]
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.pen_setPenSizeTo(value)
                
    def hinterlasse_abdruck(self):
        """hinterlasse Abdruck

        Translation string: hinterlasse Abdruck
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.pen_stamp()
                
    def pystage_addcostume(self, name, center_x=None, center_y=None, factor=1):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        center_x : FILL
        center_y : FILL
        factor : FILL
        

        Returns
        -------

        """
        return self._core.pystage_addcostume(name, center_x, center_y, factor)
                
    def pystage_addsound(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_addsound(name)
                
    def pystage_insertcostume(self, index, name, center_x=None, center_y=None, factor=1):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        index : FILL
        name : FILL
        center_x : FILL
        center_y : FILL
        factor : FILL
        

        Returns
        -------

        """
        return self._core.pystage_insertcostume(index, name, center_x, center_y, factor)
                
    def pystage_makevariable(self, name, all_sprites=True):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        all_sprites : FILL
        

        Returns
        -------

        """
        return self._core.pystage_makevariable(name, all_sprites)
                
    def pystage_replacecostume(self, index, name, center_x=None, center_y=None, factor=1):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        index : FILL
        name : FILL
        center_x : FILL
        center_y : FILL
        factor : FILL
        

        Returns
        -------

        """
        return self._core.pystage_replacecostume(index, name, center_x, center_y, factor)
                
    def pystage_setmonitorposition(self, name, x, y):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        x : FILL
        y : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setmonitorposition(name, x, y)
                
    def pystage_setmonitorstyle_large(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setmonitorstyle_large(name)
                
    def pystage_setmonitorstyle_normal(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setmonitorstyle_normal(name)
                
    def pystage_setmonitorstyle_slider(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.pystage_setmonitorstyle_slider(name)
                
    def antwort(self):
        """Antwort

        Translation string: Antwort
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_answer()
                
    def frage_und_warte(self, question):
        """frage %1 und warte

        Translation string: frage %1 und warte
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        question : FILL
        

        Returns
        -------

        """
        return self._core.sensing_askandwait(question)
                
    def farbe_berührt(self, sprite_color, color):
        """Farbe %1 berührt %2?

        Translation string: Farbe %1 berührt %2?
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite_color : FILL
        color : FILL
        

        Returns
        -------

        """
        return self._core.sensing_coloristouchingcolor(sprite_color, color)
                
    def datum_im_moment(self):
        """Datum im Moment

        Translation string: Datum im Moment
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_date()
                
    def wochentag_im_moment(self):
        """Wochentag im Moment

        Translation string: Wochentag im Moment
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_dayofweek()
                
    def stunde_im_moment(self):
        """Stunde im Moment

        Translation string: Stunde im Moment
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_hour()
                
    def minute_im_moment(self):
        """Minute im Moment

        Translation string: Minute im Moment
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_minute()
                
    def monat_im_moment(self):
        """Monat im Moment

        Translation string: Monat im Moment
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_month()
                
    def sekunde_im_moment(self):
        """Sekunde im Moment

        Translation string: Sekunde im Moment
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_second()
                
    def jahr_im_moment(self):
        """Jahr im Moment

        Translation string: Jahr im Moment
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_current_year()
                
    def tage_seit(self):
        """Tage seit 2000

        Translation string: Tage seit 2000
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_dayssince2000()
                
    def entfernung_von_mauszeiger(self):
        """Entfernung von Mauszeiger

        Translation string: Entfernung von Mauszeiger
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_distanceto_pointer()
                
    def entfernung_von(self, sprite):
        """Entfernung von %1

        Translation string: Entfernung von %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_distanceto_sprite(sprite)
                
    def taste_gedrückt(self, key):
        """Taste %1 gedrückt?

        Translation string: Taste %1 gedrückt?
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        key : FILL
        

        Returns
        -------

        """
        return self._core.sensing_keypressed(key)
                
    def lautstärke(self):
        """Lautstärke

        Translation string: Lautstärke
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_loudness()
                
    def maustaste_gedrückt(self):
        """Maustaste gedrückt?

        Translation string: Maustaste gedrückt?
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_mousedown()
                
    def maus_x_position(self):
        """Maus x-Position

        Translation string: Maus x-Position
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_mousex()
                
    def maus_y_position(self):
        """Maus y-Position

        Translation string: Maus y-Position
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_mousey()
                
    def bühnenbildname_von(self, stage='_stage_'):
        """Bühnenbildname von %2

        Translation string: Bühnenbildname von %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        stage : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_backdropname(stage)
                
    def bühnenbildnummer_von(self, stage='_stage_'):
        """Bühnenbildnummer von %2

        Translation string: Bühnenbildnummer von %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        stage : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_backdropnumber(stage)
                
    def kostümname_von(self, sprite):
        """Kostümname von %2

        Translation string: Kostümname von %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_costumename(sprite)
                
    def kostümnummer_von(self, sprite):
        """Kostümnummer von %2

        Translation string: Kostümnummer von %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_costumenumber(sprite)
                
    def richtung_von(self, sprite):
        """Richtung von %2

        Translation string: Richtung von %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_direction(sprite)
                
    def größe_von(self, sprite):
        """Größe von %2

        Translation string: Größe von %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_size(sprite)
                
    def von(self, variable, sprite='_stage_'):
        """%1 von %2

        Translation string: %1 von %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        variable : FILL
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_variable(variable, sprite)
                
    def lautstärke_von(self, sprite='_stage_'):
        """Lautstärke von %2

        Translation string: Lautstärke von %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_volume(sprite)
                
    def x_position_von(self, sprite):
        """x-Position von %2

        Translation string: x-Position von %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_xposition(sprite)
                
    def y_position_von(self, sprite):
        """y-Position von %2

        Translation string: y-Position von %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_of_yposition(sprite)
                
    def setze_stoppuhr_zurück(self):
        """setze Stoppuhr zurück

        Translation string: setze Stoppuhr zurück
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_resettimer()
                
    def setze_ziehbarkeit_auf_ziehbar(self):
        """setze Ziehbarkeit auf ziehbar

        Translation string: setze Ziehbarkeit auf ziehbar
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_setdragmode_draggable()
                
    def setze_ziehbarkeit_auf_nicht_ziehbar(self):
        """setze Ziehbarkeit auf nicht ziehbar

        Translation string: setze Ziehbarkeit auf nicht ziehbar
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_setdragmode_notdraggable()
                
    def stoppuhr(self):
        """Stoppuhr

        Translation string: Stoppuhr
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_timer()
                
    def wird_farbe_berührt(self, color):
        """wird Farbe %1 berührt?

        Translation string: wird Farbe %1 berührt?
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        color : FILL
        

        Returns
        -------

        """
        return self._core.sensing_touchingcolor(color)
                
    def wird_rand_berührt(self):
        """wird Rand berührt?

        Translation string: wird Rand berührt?
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_touchingobject_edge()
                
    def wird_mauszeiger_berührt(self):
        """wird Mauszeiger berührt?

        Translation string: wird Mauszeiger berührt?
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_touchingobject_pointer()
                
    def wird_berührt(self, sprite):
        """wird %1 berührt?

        Translation string: wird %1 berührt?
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        sprite : FILL
        

        Returns
        -------

        """
        return self._core.sensing_touchingobject_sprite(sprite)
                
    def benutzername(self):
        """Benutzername

        Translation string: Benutzername
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sensing_username()
                
    def ändere_effekt_aussteuern_links_rechts_um(self, value):
        """ändere Effekt Aussteuern links/rechts um %2

        Translation string: ändere Effekt Aussteuern links/rechts um %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_changeeffectby_pan(value)
                
    def ändere_effekt_höhe_um(self, value):
        """ändere Effekt Höhe um %2

        Translation string: ändere Effekt Höhe um %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_changeeffectby_pitch(value)
                
    def ändere_lautstärke_um(self, value):
        """ändere Lautstärke um %1

        Translation string: ändere Lautstärke um %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_changevolumeby(value)
                
    def schalte_klangeffekte_aus(self):
        """schalte Klangeffekte aus

        Translation string: schalte Klangeffekte aus
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sound_cleareffects()
                
    def spiele_klang(self, name, loop=0):
        """spiele Klang %1

        Translation string: spiele Klang %1
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        loop : FILL
        

        Returns
        -------

        """
        return self._core.sound_play(name, loop)
                
    def spiele_klang_ganz(self, name):
        """spiele Klang %1 ganz

        Translation string: spiele Klang %1 ganz
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL
        

        Returns
        -------

        """
        return self._core.sound_playuntildone(name)
                
    def setze_effekt_aussteuern_links_rechts_auf(self, value):
        """setze Effekt Aussteuern links/rechts auf %2

        Translation string: setze Effekt Aussteuern links/rechts auf %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_seteffectto_pan(value)
                
    def setze_effekt_höhe_auf(self, value):
        """setze Effekt Höhe auf %2

        Translation string: setze Effekt Höhe auf %2
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_seteffectto_pitch(value)
                
    def setze_lautstärke_auf(self, value):
        """setze Lautstärke auf %1%

        Translation string: setze Lautstärke auf %1%
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        value : FILL
        

        Returns
        -------

        """
        return self._core.sound_setvolumeto(value)
                
    def stoppe_alle_klänge(self):
        """stoppe alle Klänge

        Translation string: stoppe alle Klänge
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sound_stopallsounds()
                
    def lautstärke(self):
        """Lautstärke

        Translation string: Lautstärke
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        

        Returns
        -------

        """
        return self._core.sound_volume()
                
    def setze_name(self, name):
        """

        Translation string: 
        Engl. Translation for your reference: ...
        Engl. Documentation when available...

        
        Parameters
        ----------
        name : FILL


        Returns
        -------

        """
        return self._core.pystage_setname(name)



import pygame
import enum
import math
from pystage.core._sensing import _Sensing
from pystage.core._looks import _LooksSprite
from pystage.core.assets import CostumeManager
import random

def _deg2rad(deg):
    return deg / 360 * 2 * math.pi


def _rad2deg(rad):
    return rad / (2 * math.pi) * 360

class _Motion(_Sensing):
    def __init__(self):
        super().__init__()
        # The direction is internally stored as common angle in degrees
        # (counter-clockwise, 0 to the right)
        # Scratch uses a clockwise angle with 0 pointing to top
        self._direction = 0
        # Position is stored internally as PyGame coordinates
        # and refers to the center of the costume.
        self._pos = pygame.Vector2(0, 0)
        self.draggable = False
        self.motion_setx(0)
        self.motion_sety(0)

    def motion_turnleft(self, deg):
        self._direction += deg


    def motion_turnright(self, deg):
        self._direction -= deg


    def motion_movesteps(self, steps):
        dx = steps * math.cos(_deg2rad(self._direction))
        dy = steps * math.sin(_deg2rad(self._direction))
        self._pos += pygame.Vector2(dx, -dy) # PyGame has inverted y axis


    def motion_goto_random(self):
        half_width = int(self.stage.width/2)
        half_height = int(self.stage.height/2)
        x = random.randint(-half_width, half_width)
        y = random.randint(-half_height, half_height)
        return self.motion_gotoxy(x, y)

    motion_goto_random.opcode = "motion_goto"
    motion_goto_random.param = "TO"
    motion_goto_random.value = "_random_"


    def motion_goto_pointer(self):
        self.motion_gotoxy(self.sensing_mousex(), self.sensing_mousey())

    motion_goto_pointer.opcode = "motion_goto"
    motion_goto_pointer.param = "TO"
    motion_goto_pointer.value = "_mouse_"


    def motion_goto_sprite(self, sprite):
        self.motion_gotoxy(sprite.x_position(), sprite.y_position())

    motion_goto_sprite.opcode = "motion_goto"


    def motion_gotoxy(self, x, y):
        self.motion_setx(x)
        self.motion_sety(y)


    def motion_glideto_random(self, secs):
        half_width = int(self.stage.width/2)
        half_height = int(self.stage.height/2)
        x = random.randint(-half_width, half_width)
        y = random.randint(-half_height, half_height)
        return self.motion_glidesecstoxy(secs, x, y)

    motion_glideto_random.opcode = "motion_glideto"
    motion_glideto_random.param = "TO"
    motion_glideto_random.value = "_random_"
    motion_glideto_random.position = "%2"


    def motion_glideto_pointer(self, secs):
        self.motion_glidesecstoxy(secs, self.sensing_mousex(), self.sensing_mousey())

    motion_glideto_pointer.opcode = "motion_glideto"
    motion_glideto_pointer.param = "TO"
    motion_glideto_pointer.value = "_mouse_"
    motion_glideto_pointer.position = "%2"


    def motion_glideto_sprite(self, secs, sprite):
        return self.motion_glidesecstoxy(secs, sprite.x, sprite.y)

    motion_glideto_sprite.opcode = "motion_glideto"


    def motion_glidesecstoxy(self, secs, x, y):
        self.code_manager.current_block.gliding_seconds = secs
        self.code_manager.current_block.add_to_wait_time = secs
        self.code_manager.current_block.gliding_start_position = (self.motion_xposition(), self.motion_yposition())
        self.code_manager.current_block.gliding_end_position = (x, y)
        self.code_manager.current_block.gliding = True

    def motion_pointindirection(self, direction):
        # Scratch uses clock-wise directions with 0 to the top
        # We have anti-clockwise with 0 to the right
        self._direction = 90 - direction

    def motion_pointtowards_pointer(self):
        dx = self.sensing_mousex() - self.motion_xposition()
        dy = self.sensing_mousey() - self.motion_yposition()
        self._direction = _rad2deg(math.atan2(dy, dx))

    motion_pointtowards_pointer.opcode = "motion_pointtowards"
    motion_pointtowards_pointer.param = "TOWARDS"
    motion_pointtowards_pointer.value = "_mouse_"

    def motion_pointtowards_sprite(self, sprite):
        dx = sprite.motion_xposition() - self.motion_xposition()
        dy = sprite.motion_yposition() - self.motion_yposition()
        self._direction = _rad2deg(math.atan2(dy, dx))

    motion_pointtowards_sprite.opcode = "motion_pointtowards"


    def motion_changexby(self, value):
        self._pos.x += value


    def motion_setx(self, value):
        self._pos.x = value + self.stage.center_x


    def motion_changeyby(self, value):
        self._pos.y -= value # inverted y axis


    def motion_sety(self, value):
        self._pos.y = -value + self.stage.center_y


    def motion_ifonedgebounce(self):
        if self.rect.left < 0:
            self._direction = 180 - self._direction
            self.costume_manager.update_sprite_image()
            self._pos.x -= self.rect.left
        elif self.rect.right > self.stage.width:
            self._direction = 180 - self._direction
            self.costume_manager.update_sprite_image()
            self._pos.x -= self.rect.right - self.stage.width
        elif self.rect.top < 0:
            self._direction = -self._direction
            self.costume_manager.update_sprite_image()
            self._pos.y -= self.rect.top
        elif self.rect.bottom > self.stage.height:
            self._direction = -self._direction
            self.costume_manager.update_sprite_image()
            self._pos.y -= self.rect.bottom - self.stage.height


    def motion_setrotationstyle_leftright(self):
        self.costume_manager.rotation_style = CostumeManager.LEFT_RIGHT

    motion_setrotationstyle_leftright.opcode = "motion_setrotationstyle"
    motion_setrotationstyle_leftright.param = "STYLE"
    motion_setrotationstyle_leftright.value = "left-right"
    motion_setrotationstyle_leftright.translation="looks_effect_brightness"

    def motion_setrotationstyle_dontrotate(self):
        self.costume_manager.rotation_style = CostumeManager.NO_ROTATION

    motion_setrotationstyle_dontrotate.opcode = "motion_setrotationstyle"
    motion_setrotationstyle_dontrotate.param = "STYLE"
    motion_setrotationstyle_dontrotate.value = "don't rotate"

    def motion_setrotationstyle_allaround(self):
        self.costume_manager.rotation_style = CostumeManager.ALL_AROUND

    motion_setrotationstyle_allaround.opcode = "motion_setrotationstyle"
    motion_setrotationstyle_allaround.param = "STYLE"
    motion_setrotationstyle_allaround.value = "all around"


    def motion_xposition(self):
        return self._pos.x - self.stage.center_x

    motion_xposition.return_type = float


    def motion_yposition(self):
        return -self._pos.y + self.stage.center_y

    motion_yposition.return_type = float


    def motion_direction(self):
        # Scratch always keeps angles between -180 and 180
        dir = (90 - self._direction) % 360
        if dir <= 180:
            return dir
        else:
            return dir - 360

    motion_direction.return_type = float



import os
# from pystage.util import stderr_redirector
import sys
import io
import pygame
import pkg_resources
import pystage

_round = lambda v: pygame.Vector2(round(v.x), round(v.y))

class CostumeManager():

    ALL_AROUND = 1
    LEFT_RIGHT = 2
    NO_ROTATION = 3

    def __init__(self, owner):
        self.owner = owner
        self.costumes = []
        self.current_costume = -1
        self.rotation_style = CostumeManager.ALL_AROUND


    def add_costume(self, name, center_x=None, center_y=None, factor=1):
        if isinstance(name, str):
            costume = Costume(self, name, center_x, center_y, factor)
            self.costumes.append(costume)
            if self.current_costume==-1:
                self.current_costume = len(self.costumes) - 1
                self.update_sprite_image()
        else:
            for n in name:
                self.add_costume(n)


    def replace_costume(self, index, name, center_x=None, center_y=None, factor=1):
        costume = Costume(self, name, center_x, center_y, factor)
        del self.costumes[index]
        self.costumes.insert(index, costume)
        self.update_sprite_image()


    def insert_costume(self, index, name, center_x=None, center_y=None, factor=1):
        costume = Costume(self, name, center_x, center_y, factor)
        self.costumes.insert(index, costume)
        self.update_sprite_image()

    def next_costume(self):
        if self.current_costume == -1:
            return
        self.current_costume += 1
        if self.current_costume == len(self.costumes):
            self.current_costume = 0
        self.update_sprite_image()

    def switch_costume(self, name):
        for i, costume in enumerate(self.costumes):
            if costume.name.lower().strip() == name.lower().strip():
                self.current_costume = i
                self.update_sprite_image()
                return

    def update_sprite_image(self):
        if isinstance(self.owner, pystage.core.CoreStage):
            return
        image, new_center = self.rotate_and_scale()
        image.set_alpha((100-self.owner.ghost)/100*255)
        self.owner.image = image
        self.owner.mask = pygame.mask.from_surface(image)
        self.owner.rect = image.get_rect()
        self.owner.rect.topleft = _round(self.owner._pos) - new_center
        new_center = _round(new_center)
        if self.owner.stage.show_sprite_boundaries:
            image.blit(self.owner.mask.to_surface(), (0,0))
            pygame.draw.rect(image, "red", image.get_rect(), 1)
            pygame.draw.line(image, "red", new_center - (10, 0), new_center + (10, 0), 1)
            pygame.draw.line(image, "red", new_center - (0, 10), new_center + (0, 10), 1)


    def get_image(self):
        if self.current_costume == -1:
            return None
        return self.costumes[self.current_costume].image


    def get_costume(self):
        if self.current_costume == -1:
            return None
        return self.costumes[self.current_costume]

    
    def get_center(self):
        if self.current_costume == -1:
            return 0, 0
        return pygame.Vector2(self.costumes[self.current_costume].center_x, self.costumes[self.current_costume].center_y)


    def rotate_and_scale(self):
        # Based on:
        # https://stackoverflow.com/questions/54462645/how-to-rotate-an-image-around-its-center-while-its-scale-is-getting-largerin-py
        # Rotation settings
        flipped = False
        if self.rotation_style == CostumeManager.ALL_AROUND:
            angle = self.owner._direction
        elif self.rotation_style == CostumeManager.NO_ROTATION:
            angle = 0
        else: # LEFT_RIGHT
            angle = 0
            flipped = True if 90 < self.owner._direction % 360 < 270 else False 

        # Zoom settings
        scale = self.owner.size / 100

        old_center = self.get_center()
        old_center.y *= -1
        center_rotated = old_center.rotate(angle)

        # Corner points of the current rect
        w, h = self.get_image().get_size()
        box        = [pygame.math.Vector2(p) for p in [(0, 0), (w, 0), (w, -h), (0, -h)]]
        box_rotate = [p.rotate(angle) for p in box]
        
        # Axis aligned bounding box
        minx = min(box_rotate, key=lambda p: p[0])[0]
        maxx = max(box_rotate, key=lambda p: p[0])[0]
        miny = min(box_rotate, key=lambda p: p[1])[1]
        maxy = max(box_rotate, key=lambda p: p[1])[1]
        topleft = pygame.Vector2(minx, maxy)

        # new center
        new_center = center_rotated - topleft
        new_center.y *= -1
        new_center *= scale

        # get a rotated image
        rotozoom_image = pygame.transform.rotozoom(self.get_image(), angle, scale)
        if flipped:
            rotozoom_image = pygame.transform.flip(rotozoom_image, True, False)


        return rotozoom_image, new_center


class Costume():
    '''
    This class handles and manages costumes and backdrops.
    '''
    def __init__(self, sprite, name, center_x=None, center_y=None, factor=1):
        self.sprite = sprite
        self.file = None
        self.name = name
        internal_folder = pkg_resources.resource_filename("pystage", "images/")
        # Add further folders for translations here
        named_folders = ["", "images/", "bilder/"]
        # Folders relative to CWD
        search_folders = named_folders.copy()
        # Folders relative to python script
        for folder in named_folders:
            search_folders.append(os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])), folder))
        # Internal default sprites
        search_folders.append(internal_folder)

        for folder in search_folders:
            for ext in ["", ".bmp", ".png", ".jpg", ".jpeg", ".gif", ".svg"]:
                if os.path.exists(f"{folder}{name}{ext}"):
                    self.file = f"{folder}{name}{ext}"
                    break
            if self.file is not None:
                break
        if self.file is None:
            self.file = pkg_resources.resource_filename("pystage", "images/zombie_idle.png")
        if self.file.endswith(".svg"):
            print("\nWARNING: SVG conversion is for convenience only")
            print("and might not work as expected. It is recommended")
            print("to manually convert to bitmap graphics (png or jpg).\n")
        self.image = pygame.image.load(self.file)
        if factor!=1:
            self.image = pygame.transform.rotozoom(self.image, 0, 1.0/factor)
        self.image = self.image.subsurface(self.image.get_bounding_rect()) 
        # The offset resulting from the image crop
        offset = pygame.Vector2(self.image.get_offset())
        self.center_x = (float(self.image.get_parent().get_width()) / 2) - offset.x if center_x is None else (float(center_x) / factor) - offset.x 
        self.center_y = (float(self.image.get_parent().get_height()) / 2) - offset.y if center_y is None else (float(center_y) / factor) - offset.y
        print(f"New costume: {name} -> {self.file}")


    def __str__(self):
        return f"{self.name} ({self.center_x}, {self.center_y})"


class SoundManager():
    def __init__(self, owner):
        self.owner = owner
        self.sounds = {}

    def add_sound(self, name):
        if isinstance(name, str):
            sound = Sound(self, name)
            self.sounds[name]=sound
        else:
            for n in name:
                self.add_sound(n)

    def get_sound(self, name):
        return self.sounds[name].sound


class Sound():
    '''
    This class handles and manages sounds.
    '''
    def __init__(self, sprite, name):
        self.name = name
        self.sprite = sprite
        self.file = None
        self.sound = None
        internal_folder = pkg_resources.resource_filename("pystage", "sounds/")
        # Add further folders for translations here
        named_folders = ["", "sounds/", "klaenge/"]
        # Folders relative to CWD
        search_folders = named_folders.copy()
        # Folders relative to python script
        for folder in named_folders:
            search_folders.append(os.path.join(os.path.dirname(os.path.abspath(sys.argv[0])), folder))
        # Internal default sprites
        search_folders.append(internal_folder)

        for folder in search_folders:
            for ext in ["", ".wav", ".ogg", ".mp3"]:
                if os.path.exists(f"{folder}{name}{ext}"):
                    self.file = f"{folder}{name}{ext}"
                    break
            if self.file is not None:
                break
        if self.file.endswith(".mp3"):
            print("WARNING: MP3 is not supported in pyStage. Use wav or ogg format.")
        elif self.file is not None:
            self.sound = pygame.mixer.Sound(self.file)


    def __str__(self):
        return f"{self.name}"



import pygame

from pystage.core.code_block import CodeBlock, CodeManager
from pystage.core.assets import CostumeManager

class BaseSprite(pygame.sprite.Sprite):
    """Base Class for Mixin Classes that need access to the stage.
    """
    

    def __init__(self):
        super().__init__()
        self.stage = self.stage if hasattr(self, "stage") else None
        self.code_manager = CodeManager(self)
        self.costume_manager = CostumeManager(self)
        self.is_clone = False
        self._core = self



from os import environ

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("pystage")
    print("PyStage version {}".format(__version__))
except PackageNotFoundError:
    # package is not installed
    __version__ = "Pre 0.1"

# Prevent PyGame support prompt
environ['PYGAME_HIDE_SUPPORT_PROMPT'] = '1'
# Do not mess with the compositor
environ['SDL_VIDEO_X11_NET_WM_BYPASS_COMPOSITOR'] = '0'




import pygame
import pkg_resources


# Mixins
from pystage.core._events import _Events
from pystage.core._motion import _Motion
from pystage.core._sensing import _Sensing, _SensingSprite
from pystage.core._looks import _LooksSprite
from pystage.core._pen import _Pen
from pystage.core._variables import _Variables
from pystage.core._operators import _Operators
from pystage.core._control import _ControlSprite
from pystage.core._sound import _Sound


class CoreSprite(_Motion, _Events, _LooksSprite, _Sound, _Sensing, _SensingSprite, _ControlSprite, _Operators, _Variables, _Pen):

    def __init__(self, stage, costume="default"):
        self.stage = stage
        # Above attributes need to be set first so that mixins can access them properly
        super().__init__()
        default_file = pkg_resources.resource_filename("pystage", "images/zombie_idle.png")
        self.image = pygame.image.load(default_file)
        self.rect = self.image.get_rect()
        self.mask = pygame.mask.from_surface(self.image)
        if costume:
            self.pystage_addcostume(costume)
            self.name = "Sprite" if costume=="default" else costume 
        else:
            self.name = "Sprite"
        # The facade is the translated API
        self.facade = None
        # Make sure that the core sprite can always be accessed via _core
        self._core = self


    def pystage_setname(self, name):
        self.name = name


    def update(self, dt):
        self.code_manager._update(dt)
        self.costume_manager.update_sprite_image()
        self._update_pen()

    def __str__(self):
        return self.name



from pystage.core.sprite import CoreSprite
from pystage.core.stage import CoreStage
from pystage.core.monitors import Monitor

stage = CoreStage()
stage.pystage_addbackdrop("grid")
sprite = stage.pystage_createsprite()

# This is not yet how it should look like for end users. Either we have a
# generic way to add monitors to a stage or we need to create an API to get the
# monitors for non-variable monitors.

monitor = Monitor(stage, "Zombie: x position", color=Monitor.MOTION_COLOR)
monitor.set_function(sprite.motion_xposition)
monitor.set_position(-100, 130)


def do_something(self: CoreSprite):
    self.pystage_makevariable("test")
    self.data_setvariableto("test", 99999999999)
    self.data_showvariable("test")
    self.pystage_setmonitorposition("test", -100, 100)
    self.pystage_setmonitorstyle_large("test")
    self.motion_glidesecstoxy(5, 100, 0)



sprite.event_whenflagclicked(do_something)



stage.pystage_play()
